# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela


import numpy as np
from numpy.core.multiarray import ndarray

from ortools.linear_solver import pywraplp

from metku.optimization.solvers.optsolver import OptSolver

class SLP(OptSolver):
    """ Solver class for Sequential Linear Programming Method (SLP) """
    move_limits: np.ndarray

    def __init__(self,
                 move_limits=(0.05, 0.05),
                 gamma=1e-2,
                 C=2e5,
                 beta=10,
                 update_beta=True,
                 move_limit_type='range'):
        """ Constructor 
            :param move_limits: (tuple) relative lower and upper move limits
            :param gamma: parameter for reducing the relative move limits at each iteration
            :param C: penalty factor for safe-guarding againts potential infeasibility of LP problems
            :param beta: upper bound for the feasibility variable 'beta'. If beta = 0, then the feasibility variable
                        is ignored
            :param move_limit_type: type of move limits. 
                        Possible values: 
                            'range' .. move limits are taken with respect to the entire range of variable values. 
                            'relative' .. move limits are taken with respect to the current variable values.
        """
        super().__init__()
        self.update_beta = update_beta
        self.move_limits = np.asarray(move_limits)
        self.move_limits_hist = [np.asarray(move_limits)]
        self.move_limit_type = move_limit_type
        self.gamma = gamma
        self.C = C
        self.beta = beta
        
        self.include_beta = True
        
        if beta == 0:
            self.include_beta = False

    def take_action(self):
        """
        Defines action to take
        :return:
        """
        A, B, df, fx = self.problem.linearize(self.X)
        # bounds = [x*self.move_limits for x in self.X]
        #
        # res = linprog(df, A, B, bounds=bounds)

        solver = pywraplp.Solver('SLP',
                                 pywraplp.Solver.CLP_LINEAR_PROGRAMMING)

        # Number of constraints
        n = A.shape[0]
        # Number of variables
        m = A.shape[1]

        x = {}
        # Variables
        for i, var in enumerate(self.problem.vars):
            # Delta is the range of variable values
            # First, lb and ub are the allowable changes in the variable value
            if self.move_limit_type == "range":
                delta = var.ub - var.lb        
                #lb, ub = self.move_limits * delta
            elif self.move_limit_type == 'relative':
                delta = var.value
                #lb = self.move_limits[0]*var.value
                #ub = self.move_limits[1]*var.value
            
            lb, ub = self.move_limits * delta
            """ Set variable bounds, taking into account true upper and lower bounds """
            lb = max(var.lb, var.value - lb)
            ub = min(var.ub, var.value + ub)
            x[i] = solver.NumVar(lb,
                                 ub,
                                 var.name)                     
        # Beta        
        if self.include_beta:
            beta = solver.NumVar(0, self.beta, 'beta')
        # Constraints
        for i in range(n):
            if self.include_beta:
                solver.Add(solver.Sum([A[i, j] * x[j] for j in range(m)]) <= B[i] + beta)
            else:
                solver.Add(solver.Sum([A[i, j] * x[j] for j in range(m)]) <= B[i])

        # Objective
        if self.include_beta:
            solver.Minimize(solver.Sum([df[j] * x[j] for j in range(m)]) + self.C * beta)
        else:
            solver.Minimize(solver.Sum([df[j] * x[j] for j in range(m)]))


        sol = solver.Solve()


        # If solution if infeasible
        if sol == 2:
            print("Solution found was infeasible!")
            # Return something != 0, otherwise iteration will
            # stop because two consecutive iterations produce
            # too similar results
            return np.zeros_like(self.X)

        if self.include_beta and beta.solution_value() < 1 and self.update_beta:
            self.beta = beta.solution_value()

        X = []
        for i in range(len(x)):
            X.append(x[i].solution_value())
        X = np.asarray(X, dtype=np.float64)

        return X - self.X

    def step(self, action):

        self.C += self.C * self.gamma
        """ Reduce move limits by 'gamma'x100 percent"""        
        self.move_limits -= self.move_limits * self.gamma        
        self.move_limits_hist.append(list(self.move_limits))
        #print(self.X,action)
        self.X += action
        for i in range(len(self.X)):
            self.X[i] = np.clip(self.X[i], self.problem.vars[i].lb,
                                self.problem.vars[i].ub)

        self.problem.substitute_variables(self.X)

        return self.X.copy(), 1, False, 'INFO'

if __name__ == '__main__':
    from metku.optimization.benchmarks import *
    problem = ThreeBarTruss("continuous")
    #problem = FifteenBarTruss(prob_type='continuous')
    x0 = [var.ub for var in problem.vars]
    solver = SLP(move_limits=[0.25, 0.25], gamma=1e-3)
    fopt, xopt = solver.solve(problem, x0=x0, maxiter=100, log=False, verb=True, plot=False)
    problem(xopt)
    # #
    # # import matplotlib.pyplot as plt
    # #
    # # plt.plot(solver.fvals)
    # # plt.show()

    # problem = OptimizationProblem("Quadratic Problem")
    # problem.prob_type = 'continuous'
    # problem.obj = lambda x: x[0] ** 2 + x[1] ** 2
    # var1 = Variable("X1", 0, 5)
    # var2 = Variable("X2", 0, 5)
    # problem.add_variables([var1, var2])
    # con1 = NonLinearConstraint(lambda x: x[0] ** 2 / 20 - x[1] + 1)
    # con2 = NonLinearConstraint(lambda x: x[1] ** 2 / 20 - x[0] + 1)
    # problem.add_constraints([con1, con2])
    # solver = SLP()
    # x0 = [0.1, 0.1]
    # solver.solve(problem, x0=x0, maxiter=30)




# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Sat Apr  3 20:51:03 2021

Discrete truss optimization by mixed-integer linear programming.

TODO:
    1) Modify equilibrium equations for multiple loading conditions
    2) Kinematic stability constraints for topology optimization (auxiliary load case)
    3) Chains for topology optimization
    4) Member grouping (sizing and topology)
    5) Profile counting
290
@author: kmela
"""

import numpy as np
import re


from metku.frame2d.frame2d import SteelBeam
from metku.frame2d.simple_truss import SimpleTruss, three_bar_truss, ten_bar_truss
from metku.sections.steel import SteelSection
import metku.optimization.structopt as sop
from metku.sections.steel.catalogue import ipe_profiles, shs_profiles, hea_profiles
from metku.sections.steel import ISection, WISection, RHS, SHS
from metku.optimization.solvers.lp import LP
from metku.optimization.solvers.milp import MILP

    

class TrussMILP:
    """ Class for mixed-integer linear programming formulation for truss structures
    """
    
    def __init__(self,name,
                 truss,profiles,ulb=1e-3,uub=1e3,
                 scaling={'N':1.0, 'u':1.0},
                 groups=None,
                 member_profiles=None):
        """ Constructor
            Parameters:
            ----------
            :param name: name of the problem
            :param truss: truss structure to be optimized (ground structure for topology optimization)
            :param profiles: list of profiles from which the optimum is sought
            :param ulb: lower bound for nodal displacements
            :param uub: upper bound for nodal displacements
            :param scaling: scaling factors for axial force and displacements (lengths)
            :param groups: member groups
            :param member_profiles: for each truss member, a subset of the list 'profiles' can be given to limit
                                    the available profiles for that member.
        """
        self.name = name
        self.structure = truss
        self.profiles = profiles
        
        self.ulb = ulb*scaling['u']
        self.uub = uub*scaling['u']
        
        # Scaling
        self.scaling = scaling
        self.scaling['A'] = self.scaling['u']**2
        self.scaling['E'] = self.scaling['N']/(self.scaling['u']**2)
        
        # Design variables
        self.vars = {'profile':[[] for _ in range(self.nmem)],
                     'force':[[[] for _ in range(self.nlc)] for _ in range(self.nmem)],
                     'disp':[[] for _ in range(self.nlc)],'member':[],'node':[],'group':[]}
        
        self.__nvars = 0        
        
        self.milp = None
        
        # Resistance of cross-sections
        self.NRd = [np.floor(p.NRd)*self.scaling['A'] for p in profiles]
    
        # Member data
        self.members = [{'L':0.0,'k0':0.0,'profiles':[],
                         'NRd':[],'NbRd':[],'Nlb':[],'Nub':[],'w':[]} for _ in range(self.nmem)]
        
        # Node data:
        # keep track of supported nodes and connectivity. This data is needed
        # in topology optimization.
        self.nodes = [{'supported':0,'members':[]} for _ in range(self.nnodes)]
        
        # Allocate available member profiles
        # Assume initially that all members can have all profiles
        self.member_profiles = [list(range(self.nprof))] * self.nmem
        
        """ If member groups are given, allow for each member of a given group
            the indicated profiles
        """
        self.groups = groups
        
        if groups is not None:
            """ groups is a dict with keys: 'members','profiles' """
            for group in groups:
                for mem in group['members']:
                    self.member_profiles[mem] = group['profiles']
        
        if member_profiles is not None:
            """ member_profiles is list of dicts with keys: 'member', 'profiles' """
            for mem in member_profiles:
                self.member_profiles[mem['member']] = mem['profiles']
        
        # If no selection is given for member profiles, initialize all profiles to all members
        #if member_profiles is None:
        #    self.member_profiles = [list(range(self.nprof))] * self.nmem
        #else:
            # If member profiles is given, it must be a list of lists,
            # with length self.nmem. For each member, a list of available profile
            # indices is given.
        #    self.member_profiles = member_profiles
        
        # Groups here
        # group is a dict with keys 'members', 'profiles', which are lists
        # these lists contain indices of members that belong to the same group
        # and the corresponding profile alternatives.
        
        # Initialize member profiles
        for mem, profs in zip(self.members,self.member_profiles):
            mem['profiles'] = [self.profiles[i] for i in profs]
            mem['Nlb'] = np.zeros(len(profs))
            mem['Nub'] = np.zeros(len(profs))
            mem['NRd'] = [self.NRd[i] for i in profs]
            mem['NbRd'] = np.zeros(len(profs))
        
        #self.NbRd = np.zeros([self.nprof,self.nmem])
        
        B = self.statics_matrix
        # Initialize member data and variables that are common in all problems
        for i, mem in self.structure.members.items():
            # Calculate length and "axial stiffness"
            self.members[i]['L'] = mem.length*self.scaling['u']
            self.members[i]['k0'] = mem.E/self.members[i]['L']*self.scaling['E']
            # For each profile of the member:
            # i) calculate buckling resistance
            # ii) create profile selection variable
            # iii) calculate force variable bounds
            nlb = np.sum(np.maximum(B[:,i],0)*ulb) + np.sum(np.minimum(B[:,i],0)*uub)
            nub = np.sum(np.minimum(B[:,i],0)*ulb) + np.sum(np.maximum(B[:,i],0)*uub)
            for j, p in enumerate(self.members[i]['profiles']):
                mem.profile = p
                # Calculate buckling resistance for section 'p'
                self.members[i]['NbRd'][j] = np.floor(mem.NbRd[0])*self.scaling['N']
                
                # Add profile selection variable
                self.vars['profile'][i].append(sop.BinaryVariable("y({0:.0f},{1:.0f})".format(i,j),
                                                section=p,
                                                target={"objects":[mem],"property":"profile"}))
                self.__nvars += 1
                # Force variable bounds
                self.members[i]['Nlb'][j] = self.members[i]['k0']*p.A*nlb*self.scaling['A']
                self.members[i]['Nub'][j] = self.members[i]['k0']*p.A*nub*self.scaling['A']
        
                # Weight of member for profile 'p'
                self.members[i]['w'].append(p.weight()/self.scaling['u']*self.members[i]['L'])
                
        """ Design variables: member forces 
        Force variables are self.vars['force'][i][k][j], where i is member, j is profile and k is load case
    
        """
        for k, lc in enumerate(self.structure.load_cases):
                        
            for i, mem in self.structure.members.items():                            
                for j, p in enumerate(self.members[i]['profiles']):
                    self.vars['force'][i][k].append(
                        sop.Variable(
                            name="N({0:.0f},{1:.0f},{2:.0f})".format(i,j,lc),
                            lb=self.members[i]['Nlb'][j],
                            ub=self.members[i]['Nub'][j],
                            target=None,value=1))
                    self.__nvars += 1
                    #c.append(0)
            """ Design variables: nodal displacements 
                u[j,k]
            """
            for j in range(self.ndof):
                self.vars['disp'][k].append(
                    sop.Variable(name="u({0:.0f},{1:.0f})".format(j,k),
                                 lb=ulb,ub=uub,target=None,value=0.0))
                self.__nvars += 1
                #c.append(0)
        
        
    @property
    def nmem(self):
        """ Number of truss members """
        
        return len(self.structure.members)

    @property
    def nnodes(self):
        """ Number of truss nodes """
        return len(self.structure.nodes)

    @property
    def nprof(self):
        """ Number of profiles """
        
        return len(self.profiles)
    
    @property
    def ndof(self):
        """ number of degrees of freedom """
        return self.structure.f.nfree_dofs
    
    @property
    def nlc(self):
        """ Number of load cases """
        return len(self.structure.load_cases)
    
    @property
    def nvars(self):
        """ Number of variables """
        return self.__nvars
    
    @property
    def nprof_vars(self):
        """ Total number of profile selection variables """    
        return sum([len(y) for y in self.vars['profile']])
    
    @property
    def nstate_vars(self):
        """ Total number of state variables, i.e. the sum of force and
            displacement variables
        """
        nu = sum([len(u) for u in self.vars['disp']])
        nf = sum([len(Nlc) for Nmem in self.vars['force'] for Nlc in Nmem])
        """
        nf = 0
        for Nmem in self.vars['force']:
            for Nlc in Nmem:
                nf += len(Nlc)
        """
        return nu + nf
    
    @property
    def statics_matrix(self):
        """ Statics matrix of the truss """
        return self.structure.f.statics_matrix()
    
    def start_ndx(self,var_type):
        """ Returns the start index of the first variable of 'var_type' """
        ndx = 0
        
        if var_type == 'profiles':
            ndx = 0
        elif var_type == 'member':
            ndx = self.nprof_vars + self.nstate_vars
        elif var_type == 'node':
            ndx = self.nprof_vars + self.nstate_vars + self.nmem
        elif var_type == 'group':
            ndx = self.nprof_vars + self.nstate_vars + self.nmem + self.nnodes

        return ndx
    
    def weight_obj(self):
        """ Returns the vector for weight of the truss """
        c = np.zeros(self.nvars)
        start = 0
        # It is assumed that the profile variables are first
        # in the vector of design variables
        for mem in self.members:
            # Find the end position for the current member profile variables.
            end = start + len(mem['w'])
            # The weights of the member for each profile are stored in 'w'.
            c[start:end] = mem['w']
            start = end
        
        return c
    
    def compatibility_constraints(self,B):
        """ Constraints: compatibility conditions """
        ny_vars = self.nprof_vars    
        ndof = self.ndof
        nx = self.nvars
        Nndx = self.nprof_vars
        for k, lc in enumerate(self.structure.load_cases):
            yndx = 0
            # Starting variable index for displacement variables
            #          y            N               u 
            ustart = ny_vars + (k+1) * ny_vars + k*ndof            
            uend = ustart +ndof
            for i, mem in enumerate(self.members):                            
                for j, p in enumerate(mem['profiles']):
                    a = np.zeros(nx)
                    """ Upper bound """
                    a[yndx] = mem['Nub'][j]              # y variable
                    a[Nndx] = -1  # N variable
                    
                    a[ustart:uend] = mem['k0']*p.A*B[:,i]*self.scaling['A']                  # u variable
                    self.milp.add(sop.LinearConstraint(a,mem['Nub'][j],con_type="<", name="COMP-UB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
        
                    """ Lower bound """
                    a[Nndx] = 1  # N variable
                    a[yndx] = -mem['Nlb'][j]             # y variable
                    a[ustart:uend] *= -1    # u variable
                    self.milp.add(sop.LinearConstraint(a,-mem['Nlb'][j],con_type="<", name="COMP-LB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
                    
                    yndx += 1                
                    Nndx +=1
                    
    def strength_constraints(self):
        """ Constraints: member strength (stress constraints)
            These are expressed in terms of member forces, i.e.
            -NRd[j]*y[i,j] <= N[i,j,k] <= NRd[j]*y[i,j]
        """
        nx = self.nvars
        # Starting index for force variables
        Nndx = self.nprof_vars
        for k, lc in enumerate(self.structure.load_cases):
            yndx = 0
            for i, mem in enumerate(self.members):
                for j, Nrd in enumerate(mem['NRd']):
                    a = np.zeros(nx)
                    # Tension resistance 
                    a[yndx] = -Nrd # -Nub[j,i] #               # y variable                
                    a[Nndx] = 1  # N variable
                    self.milp.add(sop.LinearConstraint(a,0,con_type="<", name="STR-UB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
                
                    #a[i] = -slb[i]
                    # Compression resistance, including buckling 
                    a[Nndx] = -1
                    a[yndx] = -min(Nrd,mem['NbRd'][j])# Nlb[j,i] # 
                    self.milp.add(sop.LinearConstraint(a,0,con_type="<", name="STR-LB({0:.0f},{1:.0f},{2:.0f})".format(i,j,k)))
        
                    Nndx += 1
                    yndx += 1
            # Skip the displacement variables in variable indexing
            Nndx += self.ndof
            
    def add_top_opt_vars(self):
        """ Adds binary variables for topology optimization """
        # Add member and node existence variables
        for i, mem in self.structure.members.items():
            self.vars['member'].append(sop.BinaryVariable("Y({0:.0f})".format(i),
                                                          target={"objects":[mem],"property":"member"}))
            self.__nvars += 1
        
        for i, node in enumerate(self.structure.nodes):
            self.vars['node'].append(sop.BinaryVariable("z({0:.0f})".format(i),
                                                          target={"objects":[node],"property":"node"}))
            self.__nvars += 1
    
        if self.groups is not None:
            """ In case of member grouping, add corresponding binary variables """
            for g, group in enumerate(self.groups):
                for p in group['profiles']:
                    self.vars['group'].append(sop.BinaryVariable("w({0:.0f},{1:0f})".format(g,p)))
                    self.__nvars += 1
    
    
    def gather_node_data(self):
        """ Collects needed topological data for nodes.
            i) which nodes are supported
            ii) which members are connected to each node
        """
        for i, mem in self.structure.members.items():        
            for n in mem.nodes.keys():
                self.nodes[n]['members'].append(i)
        
        for i, supp in self.structure.supports.items():            
            self.nodes[supp.node_id]['supported'] = len(supp.dofs)
    
    def add_topology_constraints(self):
        """ Adds the following constraints:
            1) members attached to vanishing node vanish
            2) minimum number of members meeting at existing node            
            3) necessary condition for kinematic stability
            4) minimum number of supported nodes
            5) constraints for groups (optional)
        """
        nx = self.nvars
        
        Ystart = self.start_ndx('member')
        Zstart = self.start_ndx('node')
        # 1) Members attached to vanishing node must vanish
        for j, node in enumerate(self.nodes):            
            for m in node['members']:
                a = np.zeros(nx)
                a[Zstart+j] = 1
                a[Ystart+m] = -1
                self.milp.add(sop.LinearConstraint(a,0,con_type="<",
                                                   name="Y[{0:.0f}] <= Z[{1:.0f}]".format(m,j)))
        
        # 2) Minimum number of members meeting at existing node
        for j, node in enumerate(self.nodes):            
            a = np.zeros(nx)
            mndx = [m + Ystart for m in node['members']]
            a[mndx] = -1
            if node['supported']:
                a[Zstart+j] = 1
            else:
                a[Zstart+j] = 2
            
            self.milp.add(sop.LinearConstraint(a,0,con_type="<",
                                                   name="Sum Y[i] => C*Z[{0:.0f}]".format(j)))
        
        # 3) Necessary condition for kinematic stability
        a = np.zeros(nx)
        a[[Ystart + i for i in range(self.nmem)]] = -1
        ND = 2
        a[[Zstart + i for i in range(self.nnodes)]] = ND
        
        for j, node in enumerate(self.nodes):
            if node['supported']:
                a[Zstart + j] -= node['supported']
        
        self.milp.add(sop.LinearConstraint(a,0,con_type="<",
                                           name="Kinematic stability (necessary cond.)"))
        
        # 4) Minimum number of supported nodes is 2
        a = np.zeros(nx)
        for j, node in enumerate(self.nodes):
            if node['supported']:
                a[Zstart + j] = -1
        
        self.milp.add(sop.LinearConstraint(a,-2,con_type="<",
                                           name="At least 2 supported nodes"))
        
        # 5) Constraints for groups (if any)
        if self.groups is not None:
            Wstart = self.start_ndx('group')
            dW = 0
            for g, group in enumerate(self.groups):
                # First constraints: sum of group profile variables can be at most 1
                aw = np.zeros(nx)
                aw[[Wstart + dW + i for i in range(len(group['profiles']))]] = 1
                
                self.milp.add(sop.LinearConstraint(aw,1,con_type="<",
                                                   name="SUM W[{0:.0f},j] <= 1".format(g)))
                
                # Second constraints:
                # For all members of the group:
                # y(i,j) <= w(j) for all j
                for mem in group['members']:
                    # Find index of the first member profile variable
                    yndx = self.milp.vars.index(self.vars['profile'][mem][0])
                    
                    for j, p in enumerate(group['profiles']):
                        a = np.zeros(nx)
                        a[Wstart + dW + j] = -1
                        a[yndx+j] = 1
                        
                        self.milp.add(sop.LinearConstraint(a,0,con_type="<",
                                                   name="y({0:.0f},{1:.0f}) <= W[{0:.0f},j]".format(mem,p)))
                
                # Third constraints:
                # SUM y[i,j] >= w[j] for all j
                for j, p in enumerate(group['profiles']):
                    a = np.zeros(nx)
                    a[Wstart + dW + j] = 1
                    
                    for mem in group['members']:
                        yndx = self.milp.vars.index(self.vars['profile'][mem][j])
                        a[yndx] = -1
                    
                    self.milp.add(sop.LinearConstraint(a,0,con_type="<",
                                                   name="SUM y(i,{0:.0f}) >= W[{0:.0f},{1:.0f}]".format(j,g)))
                
                
                dW += len(group['profiles'])
                        
                
            #print(Wstart)
        
                          
    def create_milp(self,problem_type='sizing',constraints={'strength':True,'buckling':True}):
        """ Create OptimizationProblem class instance that contains
            the mixed-integer linear program of the truss.
            
             Parameters:
            ------------
            :param problem_type: 'sizing' or 'topology'
            :param constraints: different kinds of constriaints to be included in the problem
    
            :type problem_type: string
            :type constraints: dict
        """
        dvars = []
        
        # Add profile selection variables
        for y in self.vars['profile']:
            dvars.extend(y)
        
        # For each load case, append the force and displacement variables
        # Design variable vector is then
        # x = [y, N[0] u[0], N[1] u[1], ... N[nlc] N[nlc]]
        for k, disp_vars in enumerate(self.vars['disp']):
            for mem_force_var in self.vars['force']:
                dvars.extend(mem_force_var[k])
            dvars.extend(disp_vars)
        
        if problem_type == 'topology':
            self.add_top_opt_vars()
            dvars.extend(self.vars['member'])
            dvars.extend(self.vars['node'])
            
            if self.groups is not None:
                dvars.extend(self.vars['group'])
            
    
        """ Initialize optimization problem with the created variables """
        self.milp = sop.OptimizationProblem(name=self.name,variables=dvars,structure=self.structure)
    
        """ Objective function: truss weight.
            This is the same for all problem types.
        """
        obj = sop.LinearObjective("Weight", self.weight_obj())
        self.milp.add(obj)
        
        ndof = self.ndof
        nvars = self.nvars
        
        """ Constraints: unique profile selection """
        Aeq = np.zeros([self.nmem,nvars])
        for i, var in enumerate(self.milp.vars):
            """ Find the profile selection variables based on their name """
            s = var.name
            if s[0] != 'y':
                break
            else:
                """ The first number in the name of the variable corresponds to the member index """
                m = re.findall(r'\d+',s)
                Aeq[int(m[0]),i] = 1
        
        if problem_type == 'topology':
            lhs = 0.0
            Yndx = self.start_ndx('member')
        else:
            lhs = 1.0            
        for i, a in enumerate(Aeq):                    
            if problem_type == 'topology':
                a[Yndx + i] = -1
                con_name = "Sum y({0:.0f},j)=Y({0:.0f})".format(i)                
            else:
                con_name = "Sum y({0:.0f},j)=1".format(i)
                
            self.milp.add(sop.LinearConstraint(a,lhs,con_type="=", name=con_name))
        
        if problem_type == 'topology':
            self.gather_node_data()
            self.add_topology_constraints()
        
        """ Constraints: equilibrium equations """
        # Create augmented statics matrix, whose columns correspond to
        # the columns of the original statics matrix copied as many times
        # as there are profile variables for each member.        
        B0 = np.zeros([ndof,nvars])
        B = self.statics_matrix
        
        # TODO! This has to be accommodated for several load cases!
        start = self.nprof_vars
        for i, yvars in enumerate(self.vars['profile']):
            end = start + len(yvars)
            B0[:,start:end] += np.reshape(B[:,i],(ndof,1))
            start = end
                            
        for j, lc in enumerate(self.structure.load_cases):
            # Load vector for the load case, including only unsupported dofs """
            p = np.delete(self.structure.f.global_load_vector(lc),self.structure.f.supp_dofs)*self.scaling['N']
            # a is a placeholder for the coefficient vector of a linear constraint """
            a = np.zeros(nvars)
            for i, (b,pi) in enumerate(zip(B0,p)):
                # The first 'nmem' variables are the cross-sectional areas.
                #    Insert the current row of the statics matrix as the coefficient
                #    vector.            
                self.milp.add(sop.LinearConstraint(b,pi,con_type="=", name="EQ-{0:.0g}-LC{1:.0g}".format(i,lc)))
        
        # Constraints: compatibility and strength
        self.compatibility_constraints(B)
        self.strength_constraints()
                
        
        return lp
    
    def substitute_design(self,x):
        """ Substitutes the design in 'x' to the truss structure """
        
        self.milp.substitute_variables(x)
    
    def plot_design(self,x):
        """ Plots the design represented by the variable vector 'x' """
        self.substitute_design(x)
        self.milp.structure.plot()

    def plot_ground_structure(self):
        """ plots the ground structure with member numbers """
        
        for mem in self.structure.members.values():
            mem.active = True
        
        self.structure.plot(print_text=False)
        
if __name__ == '__main__':
    #t = three_bar_truss(L=3000,F1=-200e3,F2=-250e3,nlc=1)
    t = ten_bar_truss(L=3000,F=200e3)
    
    #lp = truss_lp(t,0,1e8)
    #LPsolver = LP()
    #sol = LPsolver.solve(lp,verb=True)
    
    P = []
    
    hmax = 120
    hmin = 50
    tmin = 4.0
    
    for key, val in shs_profiles.items():
        if val['h'] <= hmax and val['h'] >= hmin and val['t'] >= tmin:
            P.append(SHS(val['h'],val['t']))
    
    """
    member_profiles = [[],[],[]]
    
    member_profiles[0] = [0, 2, 4, 6, 8]
    member_profiles[1] = [3, 5, 7]
    member_profiles[2] = [0, 1, 2, 9, 10, 11]
    """
    groups = []
    groups.append({'members':[6,3],'profiles':[7,8,9,10,11]})
    groups.append({'members':[1,7],'profiles':[1,2,3,4,5,6,7]})
    
    member_profiles = None
    lp = TrussMILP('TrussMILP',t,P,scaling={'N':1e-3,'u':1e-1},ulb=-200,uub=200,member_profiles=member_profiles,groups=groups)
    
    lp.create_milp(problem_type='sizing')
    MILPsolver = MILP()
    MILPsolver.solve(lp.milp,verb=True)
    x = MILPsolver.X
    
    lp.plot_design(x)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Mon Apr  1 13:12:18 2019

@author: kmela

Contents

Class Workshop

Class CostCentre
Class BlastingCost(CostCentre)
Class CuttingCost(CostCentre)
Class PaintingCost(CostCentre)
Class AssemblingCostWeld(CostCentre)

TODO:
    Beam Welding .. Haapio 4.5.4, for welded beams
    Sawing .. Haapio 4.5.5
    Drilling .. Haapio 4.5.6
    Coping .. Haapio 4.5.7
    Fabrication of parts .. Haapio 4.5.8 (e.g. angle profiles, plates)
    Assembling (Bolts) .. Haapio 4.5.9
    Post-treatment and inspection .. Haapio 4.5.10
    Drying
    Erecting

"""

import copy
from colorama import Fore
from metku.cost.cost_data import BASIC_STEEL_PRICE

NOT_IMPLEMENTED = "NOT IMPLEMENTED YET! Add this feature: "

work_days_per_year = 252
work_hours_per_day = 8
work_minutes_per_year = work_days_per_year*work_hours_per_day*60

wages = {"plate_and_construction": 18.37,
         "welding": 18.13,
         "machine": 19.76,
         "quality_control":19.16,
         "erecting_work":18.76
         }

overhead = 0.7

wage_per_minute = copy.deepcopy(wages)

# change hourly rates to euros per minute, including overheads
for key in wage_per_minute:
    wage_per_minute[key] *= (1+overhead)/60

#total_wages_per_min = 

""" Construction of the workshop, based on
    Haahtela & Kiiras, Talonrakennuksen kustannustieto 2015
    Alue 4, Kasvukeskukset, hintataso 78
""" 
construction_cost = 1000 # €/m2

""" Cost of plot
    It is assumed that the workshop is located in Seinäjoki, Finland
"""
plot_efficiency = 0.5
raw_land_cost = 6 # €/m2
land_cost = raw_land_cost/plot_efficiency

""" Maintenance cost of real estate
    KH-cards KH X1-00244, KH X1-00291 and KH X1-00379
    Teollisuushallin ylläpito 2000/4, KH X1-00291 taulukko 6
"""
industrial_hall_maintenance = 6.385 # €/(m3*a)

""" Kiinteistön ylläpidon kustannusindeksi 
    rakennetyypeittäin (Tilastokeskus), 2000=100, Teollisuushalli
"""
maintenance_index = 181 # Q4/2014

building_height = 8.0 # m
real_estate_maintenance_cost = building_height*industrial_hall_maintenance*maintenance_index/100

#print(f'{real_estate_maintenance_cost:.2f} €')

""" Cost of electricity """
cost_energy = 0.1 # €/kWh
cost_energy_min = cost_energy/60

""" Investements 
    IEq .. interest rate of equipment
    nEq .. life of equipment (years)
    IRealS .. interest reate of real estate
    nRealS .. life of real estate (years)
"""

IEq = 0.05
nEq = 20
IRealS = 0.05
nRealS = 50

""" Paint data
    vol_solid .. volume of solids
    DFT .. dry film thickness
    price .. €/l
"""
tip_flow = 900000 # [mm3/min]

paints = {"alkyd": {"vol_solid":[0.48,0.45,0.45], "DFT":[80,40,40],"price":[2.8,3.6,3.6],"time":3,"TPP":0.0,"CP":0.0},
          "epoxy": {"vol_solid":[0.50,0.50,0.48], "DFT":[60,60,40],"price":[4.2,4.2,4.9],"time":14,"TPP":0.0,"CP":0.0},
          "polyurethane": {"vol_solid":[0.50,0.50], "DFT":[60,60],"price":[4.2,5.9],"time":27,"TPP":0.0,"CP":0.0},
          "acryl": {"vol_solid":[0.48,0.48,0.40], "DFT":[60,60,40],"price":[6.4,6.4,6.4],"time":7,"TPP":0.0,"CP":0.0},
          }

# Calculate basic values for productive time and cost of consumables
for paint, films in paints.items():    
    nfilms = len(films["vol_solid"])
    TPP = 0.0
    CP = 0.0
    for i in range(nfilms):
        tloss = 0.01+0.3*films["DFT"][i]*1e-3
        P0 = 1/films["vol_solid"][i]*(films["DFT"][i]*1e-3+tloss)
        TPP += P0
        CP += films["price"][i]*1e-6*P0

    TPP = TPP/tip_flow
    paints[paint]["TPP"] = TPP
    paints[paint]["CP"] = CP


""" Welding consumables, prices """
""" Welding process 135 (MAG with solid wire) """
wire_prices = {"S355": 1.66,
               "S420": 1.66,
               "S550": 3.5,
               "S690": 2.34}

sawing_times = {"setup":2,
                "blade_descent":0.25,
                "blade_raising":0.5,
                "profile_fixing":0.5,
                }

sawing_base_time = sawing_times["setup"]+\
            2*(sawing_times["blade_descent"]+sawing_times["blade_raising"]+\
               sawing_times["profile_fixing"])

saw_bevel_time = 1.0 # min

sawing_parameters = {"S235":{"Sm":2.0, "Q":1800},
                     "S355":{"Sm":2.0, "Q":1800},
                     "S420":{"Sm":2.0, "Q":1800},
                     "S500":{"Sm":2.0, "Q":1800},
                     "S550":{"Sm":2.0, "Q":1800},
                     "S690":{"Sm":2.0, "Q":1800},
                     "S700":{"Sm":2.0, "Q":1800}}

""" Cost of one saw blade [€] """
#saw_blade_cost = 100.0
saw_blade_cost = 170.0

def print_initial_cost_data():
    """ Prints a set of initial data used in cost calculations """
    
    header = "-----"
    
    print("Cost analysis following Haapio (2012)")
    print(header + " Working time " + header)
    print(f"Working days per year: {work_days_per_year} d")
    print(f"Working hours per shift: {work_hours_per_day} h")
    print(f"Working minutes per year: {work_minutes_per_year}")
    
    print('\n' + header + " Wages " + header)
    
    for key, value in wages.items():
        print(f'{key}: {value:.2f} €/h')
    
    print(f'Overhead: {overhead}')

    for key, value in wage_per_minute.items():
        print(f'{key}: {value:.3f} €/min')
    
    print('\n' + header + " Real estate " + header)
    
    print(f"Construction cost: {construction_cost} €/m2") 
    print(f"Plot efficiency: {plot_efficiency} ") 
    print(f"Raw land cost: {raw_land_cost} €/m2") 
    print(f"Land cost (Seinäjoki, Finland): {land_cost} €/m2") 
    print(f"Interest rate of real estate: {IRealS*100:.1f}%")
    print(f"Life time of real estate: {nRealS} a")
   
    print(f"Maintenance cost, industrial hall: {industrial_hall_maintenance} €/(m3*a)") 
    print(f"Maintenance index (Q4/2014): {maintenance_index} ") 
    print(f"Building height: {building_height} m")
    print(f"Real estate maintenance cost: {real_estate_maintenance_cost:.2f} €/(m2*a)")
    
    print('\n' + header + " Cost of electricity " + header)
    print(f'Cost of energy: {cost_energy} €/kWh')
    print(f'Cost of energy per minute: {cost_energy_min:.2f} €/kWh/min')

    print('\n' + header + " Investment of equipment " + header)
    print(f'Interest rate of equipment: {IEq*100} %')
    print(f'Life time of equipment: {nEq} a')
    
    print('\n' + header + " Painting " + header)
    for paint, values in paints.items():
        print(f"{paint}: TPP = {values['TPP']:.2e}, CP = {values['CP']:.2e} ")
        
    
    
    """
    paints = {"alkyd": {"vol_solid":[0.48,0.45,0.45], "DFT":[80,40,40],"price":[2.8,3.6,3.6],"time":3,"TPP":0.0,"CP":0.0},
              "epoxy": {"vol_solid":[0.50,0.50,0.48], "DFT":[60,60,40],"price":[4.2,4.2,4.9],"time":14,"TPP":0.0,"CP":0.0},
              "polyurethane": {"vol_solid":[0.50,0.50], "DFT":[60,60],"price":[4.2,5.9],"time":27,"TPP":0.0,"CP":0.0},
              "acryl": {"vol_solid":[0.48,0.48,0.40], "DFT":[60,60,40],"price":[6.4,6.4,6.4],"time":7,"TPP":0.0,"CP":0.0},
              }
    """
# Annual cost, Haapio, eq (2)
# Uniform series end-of-period installment [e/a]
def annual_cost(P, I, n):
    # P = a sum of money invested in the initial year less resale value [e]
    # I = interest rate (cost of capital) [%]
    # n = time, the number of units (years) over which interest accumulates

    return P*((I*(1.0 + I)**n)/((1.0 + I)**n - 1.0))       # [e/a]

class CostCentre:
    """
        Class for cost centres
    """
    def __init__(self,length=20,width=10,workers=1,
                 eq_invest=1000,eq_resale=0,eq_maint=1000,eq_year=nEq):
        
        self.name = "Cost Centre"
        self.length = length
        self.width = width
        self.nw = workers        
        
        self.times = {"productive":0.0,"non-productive":0.0}
        #self.productive_time = 0.0
        #self.non_productive_time = 0.0
        
        self.costs = {"labour":0.0,
                      "equipment":0.0,
                      "equip_maintenance":0.0,
                      "real_estate":0.0,
                      "real_maintenance":0.0,                     
                      }
        
        self.productive_costs = {"consumables":0.0,
                                 "energy":0.0
                                 }
        
        self.consumables_cost = 0.0
        self.utility = 1.0             
        
        """ Unit costs of real estate can be calculated similarly to 
            all cost centres.
            
            Other unit costs depend on the specifics.
        """
        # Investment cost of workspace area [e]
        P_invest_RealS = self.Area*construction_cost
        # Resale value of workspace [e], equals value of land
        P_resale_RealS = self.Area*land_cost          
        self.costs["real_estate"] = annual_cost(P_invest_RealS - P_resale_RealS, IRealS, nRealS)/work_minutes_per_year
        
        # Real estate maintenance cost
        c_maintenance_RealS = self.Area*real_estate_maintenance_cost
        # Real estate maintenance cost[e/min]
        self.costs["real_maintenance"] = c_maintenance_RealS/work_minutes_per_year
        
        # Equipment installment cost [e/min]
        # Investment of equipment [e]
        
        self.costs["equipment"] = annual_cost(eq_invest - eq_resale, IEq, eq_year)/work_minutes_per_year

        # Maintenance cost of equipment        
        self.costs["equip_maintenance"] = eq_maint/work_minutes_per_year
        
    @property
    def Area(self):
        return self.length*self.width
        
    # Total cost of cost center [e]
    def cost(self):
        return sum(self.times.values())*sum(self.costs.values())/self.utility + \
                self.times["productive"]*sum(self.productive_costs.values()) + \
                self.consumables_cost
    
    def info(self):
        
        print(Fore.GREEN + f"Cost centre: {self.name}")
        print(Fore.RESET + f"Area: {self.Area:.2f} m2")
        
        print(f"Labour costs: {self.costs['labour']:.2f} €/min")
        print(f"Equipment costs: {self.costs['equipment']:.2f} €/min")
        print(f"Equipment maintenance costs: {self.costs['equip_maintenance']:.2f} €/min")
        print(f"Real estate costs: {self.costs['real_estate']:.2f} €/min")
        print(f"Real estate maintenance costs: {self.costs['real_maintenance']:.2f} €/min")
# Cost center for beam welding
class BeamWeldingCost(CostCentre):

    def __init__(self,length=20,width=15,workers=2,
                 eq_invest=200000.0,eq_resale=0.0,eq_maint=1000.0):
    
        CostCentre.__init__(self, length,width,workers,
                            eq_invest,eq_resale,eq_maint)
        
        self.name = "Beam Welding"
        
        self.times["non-productive"] = 6.25 # minutes
        
        self.nwelds = 2 # number of simultaneous welds
        self.deposit_rate = 0.14 # kg/min
        self.wire_price = 1.84 # €/kg
        self.flux_price = 1.46 # €/kg
        self.flux_wire_ratio = 1.0
        

        # Labour cost
        self.costs["labour"] = self.nw*wage_per_minute["machine"] # Hourly cost [e/h]
         
                     
        # Cost of consumables [e/min] TODO!
        self.productive_costs["consumables"] = 1.0
        # Energy
        # Total power consumption of one weld head [kW]
        self.power = 24                       
        # Cost of energy [e/min]
        self.productive_costs["energy"] = self.power*cost_energy_min                

    def ProductiveTime(self,weld_length,weld_size,weld_type="fillet"):
        """ Calculate the productive time for welding
            input: Lb .. length of steel piece (mm)
            output: productive time (min)
        """
        
        
        TP = Lb/self.vc
        self.times["productive"] = TP
        return TP  


    def welding_time(self):
        """ Time required for welding [min]
            input: weld_size .. throat thickness of fillet weld or
                                depth of single-bevel butt weld
                   weld_length [mm]
                   weld_type .. "fillet" or "butt"
        """
        if self.weld_type == "fillet":
            a = self.weld_size            
            tweld = self.weld_length*1e-3*(0.4988*a**2-0.0005*a+0.0021)            
        elif self.weld_type == "butt":
            b = self.weld_size
            tweld = self.weld_length*1e-3*(0.249*b**2+0.0096*b-0.0506)
        
        return tweld

    def productive_time(self,weld_length,weld_size,weld_type="fillet"):
        """ Productive time of welding
            
        """
        # Productive time [min], painting        
        if self.weld_type == "fillet":
            a = self.weld_size
            weld_weight = self.weld_length*a**2*7850e-9
        elif self.weld_type == "butt":
            b = self.weld_size
            weld_weight = self.weld_length*b**2/2*7850e-9
        
        self.times["productive"] = weld_weight/self.deposit_rate
        

    def consumables(self):
        """ Cost of welding consumables            
        """
        
        if self.weld_type == "fillet":
            a = self.weld_size
            weld_weight = self.weld_length*a**2*7850e-9
        elif self.weld_type == "butt":
            b = self.weld_size
            weld_weight = self.weld_length*b**2/2*7850e-9
                
        self.consumables_cost = weld_weight*(self.wire_price + self.gas_consumption*self.gas_price)
                    
        return self.consumables_cost      


    
# Cost center for blasting
class BlastingCost(CostCentre):

    def __init__(self,length=40,width=10,workers=1,speed=3000,
                 eq_invest=200000.0,eq_resale=0.0,eq_maint=1000.0):
    
        CostCentre.__init__(self, length,width,workers,
                            eq_invest,eq_resale,eq_maint)
        
        self.name = "Blasting"
        # Productive time
        self.vc = speed                         # Conveyor speer [mm/m]        

        # Labour cost
        self.costs["labour"] = self.nw*wage_per_minute["machine"] # Hourly cost [e/h]
         
        # Consumables: grains
        # Grain consumption on blasting [kg/min], equipment utilization ratio 1
        # For utilization ratio of 0.5, grain consumption is 10kg/8h
        grain_consump = 0.042                   
        # Grain weight price [e/kg]
        #grain_unit_cost = 0.50                  
        grain_unit_cost = 0.75 

        # Cost of consumables [e/min]
        self.productive_costs["consumables"] = grain_consump*grain_unit_cost
        # Energy
        # Total power consumption using four blasting turbines [kW]
        P_total = 40                                    
        # Cost of energy [e/min]
        self.productive_costs["energy"] = P_total*cost_energy_min                

    def ProductiveTime(self,Lb):
        """ Calculate the productive time for steel piece of length Lb
            input: Lb .. length of steel piece (mm)
            output: productive time (min)
        """
        TP = Lb/self.vc
        self.times["productive"] = TP
        return TP

    def cost(self,Lb):
        
        self.ProductiveTime(Lb)
        
        return super().cost()

# Cost center for cutting
class CuttingCost(CostCentre):

    def __init__(self, length=20.0,width=10.0, workers=2.0,
                 eq_invest=280000.0,eq_resale=0.0,eq_maint=1000.0):

        CostCentre.__init__(self, length,width,workers,
                            eq_invest,eq_resale,eq_maint)
        
        self.name = "Cutting"
        
        self.times["non-productive"] = 3.0          
        
        # Labour cost
        self.costs["labour"] = self.nw*wage_per_minute["machine"] # Hourly cost [e/h]
            
        

    def ProductiveTime(self,t_p,cut_length):
        """ Calculate the productive time
            input: Lb .. length of steel piece (mm)
            output: productive time (min)
        """
        # Cutting speed according to thickness [mm] of the plate
        if t_p <= 30.0:   # Plasma cutting
            v_cut = (8.9212*t_p**2.0 - 486.87*t_p + 8155.8)     # [mm/min]
        else:   # Flame cutting
            v_cut = -4.1939*t_p + 658.67                        # [mm/min]
            
        TP = cut_length/v_cut                                   # [min]
        self.times["productive"] = TP
        return TP
    
    def Consumables(self,t_p):
        """ Calculate the cost of consumables
            input: t_p .. plate thickness
            
        """
        # Consumables: Gases, plasma electrodes and nozzles
        if t_p <= 30.0:   # Plasma cutting
            # Total consumables cost [e/min]
            cutting_oxygen = 0.071                                       # [m^3/min]
            oxygen_cost = 0.3                                            # [e/min]
            nozzle_unit_cost = 40                                        # [e]
            nozzle_life = 480                                            # [min]
            nozzle_cost = nozzle_unit_cost/nozzle_life                   # [e/min]
            self.productive_costs["consumables"] = cutting_oxygen*oxygen_cost + nozzle_cost
            
        else:   # Flame cutting
            # Total consumables cost [e/min]
            propane_consump = 0.0063                                        # [m^3/min]
            propane_cost = 18.40                                            # [e/m^3]
            preheat_oxygen_consump = 0.025                                  # [m^3/min]
            cutting_oxygen_consump = 1.0e-5*t_p**2.0 + 0.001*t_p + 0.0224   # [m^3/min]
            oxygen_cost = 4.18                                              # [e/m^3]
            self.productive_costs["consumables"] = propane_consump*propane_cost + (preheat_oxygen_consump + cutting_oxygen_consump)*oxygen_cost        
    
    def Energy(self,t_p):
        """ Cost of energy used by the machine """
        
        if t_p <= 30:
            # Plasma cutting
            P_total = 72.0
        else:
            # Flame cutting
            P_total = 0.0
            
        # Cost of energy [e/min]
        self.productive_costs["energy"] = P_total*cost_energy_min

    def cost(self,t_p,cut_length):
        """ Evaluate different costs and run the total cost
            calculation
        """
        self.ProductiveTime(t_p, cut_length)
        self.Consumables(t_p)
        self.Energy(t_p)
        return super().cost()

class PaintingCost(CostCentre):    

    def __init__(self, length=15.0,width=5.0, workers=1.0,
                 paint = "alkyd"):
        """ NOTE: Drying costs are missing, 26.1.2022 """

        CostCentre.__init__(self, length,width,workers,
                            eq_invest=0.0,eq_resale=0.0,eq_maint=0.0)        
        
        self.name = "Painting"
        
        self.paint = paint
        # flow of paint through the tip of the paint gun
        # (mm3/min)
        self.tip_flow = tip_flow
            
        self.costs["labour"] = self.nw*wage_per_minute["machine"]
            

    def productive_time(self,Ap):
        """ Productive time of painting
            input: Ap .. area to be painted (mm2)
        """
        # Productive time [min], painting
        T_P = paints[self.paint]["TPP"]*Ap       
        
        self.times["productive"] = T_P
        return T_P

    def consumables(self,Ap):
        """ Cost of painting consumables
            input: Ap .. area to be painted (mm2)
        """
        # Consumables: Paint
        # Total cost of non-time-related consumables [e]
        C_C = paints[self.paint]["CP"]*Ap
       
        self.consumables_cost = C_C
        
    def cost(self,Ap):
        """ Evaluate total painting cost """
        
        self.productive_time(Ap)
        self.consumables(Ap)
        
        return super().cost()
        
class AssemblingCostWeld(CostCentre):    
    """ Cost of welding an assembly """
    
    def __init__(self, length=15.0,width=5.0, workers=1.0,
                 eq_invest=5000.0,eq_resale=0.0,eq_maint=0.0,
                            eq_year=10,
                            weld_type="fillet",
                            weld_length=250.0,
                            weld_size=4.0,
                            steel_grade="S355"):
        """
            input:  length, width .. dimensions of cost centre
                    workers .. number of workers
                    eq_invest .. equipment investment cost
                    eq_resale .. equipment resale value
                    eq_maint .. equipment maintenance cost (€/year)
                    eq_year .. life of equipment (years)
                    weld_type .. "fillet" or "butt"
                    weld_length .. (mm)
                    weld_size .. (mm)
                    steel_grade .. of steel piece to be welded. This determines
                                    the cost of the wire
        """
        CostCentre.__init__(self, length,width,workers,
                            eq_invest,eq_resale,eq_maint,eq_year)                            
        
        self.name = "Assembly welding"
        
        self.weld_type = weld_type
        self.weld_length = weld_length
        self.weld_size = weld_size
        
        number_of_tacks = 3
        
        self.tacking_times = {"debur":0.5,
                         "insertion":0.5,
                         "clamp":0.27,
                         "tack":0.09*number_of_tacks+0.05}
        
        self.power = 200*30*1e-3 # current * voltage (kW)
        self.productive_costs["energy"] = self.power*cost_energy_min
        
        self.gas_consumption = 0.4  # m3/kg wire
        self.gas_price = 3.79      # euro/m3 gas        
        self.wire_price = wire_prices[steel_grade] # euro/kg wire            
        
        self.tacking_time = sum(self.tacking_times.values())
        
        self.costs["labour"] = self.nw*wage_per_minute["machine"]
        self.times["productive"] = self.welding_time()+self.tacking_time
        self.consumables()
            
    def welding_time(self):
        """ Time required for welding [min]
            input: weld_size .. throat thickness of fillet weld or
                                depth of single-bevel butt weld
                   weld_length [mm]
                   weld_type .. "fillet" or "butt"
        """
        if self.weld_type == "fillet":
            a = self.weld_size            
            tweld = self.weld_length*1e-3*(0.4988*a**2-0.0005*a+0.0021)            
        elif self.weld_type == "butt":
            b = self.weld_size
            tweld = self.weld_length*1e-3*(0.249*b**2+0.0096*b-0.0506)
        
        return tweld

    def productive_time(self):
        """ Productive time of welding
            input: Ap .. area to be painted (mm2)
        """
        # Productive time [min], painting        
        self.times["productive"] = self.welding_time()+self.tacking_time
        

    def consumables(self):
        """ Cost of welding consumables            
        """
        
        if self.weld_type == "fillet":
            a = self.weld_size
            weld_weight = self.weld_length*a**2*7850e-9
        elif self.weld_type == "butt":
            b = self.weld_size
            weld_weight = self.weld_length*b**2/2*7850e-9
                
        self.consumables_cost = weld_weight*(self.wire_price + self.gas_consumption*self.gas_price)
                    
        return self.consumables_cost
    
    def cost(self,weld_size,weld_length,weld_type='fillet'):
        """ Evaluate total painting cost """
        
        self.weld_size = weld_size
        self.weld_length = weld_length
        self.weld_type = weld_type
        
        self.consumables()
        self.productive_time()
        
        return super().cost()
        
class SawingCost(CostCentre):    
    """ Cost of sawing """
    
    def __init__(self, length=35.0,width=15.0, workers=1.0,
                 eq_invest=310e3,eq_resale=0.0,eq_maint=1e3,eq_year=20):
        """
            input:  length, width .. dimensions of cost centre
                    workers .. number of workers
                    eq_invest .. equipment investment cost
                    eq_resale .. equipment resale value
                    eq_maint .. equipment maintenance cost (€/year)
                    eq_year .. life of equipment (years)                    
        """
        CostCentre.__init__(self, length,width,workers,
                            eq_invest,eq_resale,eq_maint,eq_year)                            
        
        self.name = "Sawing"
        
        # Here, a basic value for the non-productive time is calculated.
        # It consists of setting up the machine, fixing the profile,
        # descending and ascending the blade for both ends
        self.times["non-productive"] = sawing_times["setup"]+\
            2*(sawing_times["blade_descent"]+sawing_times["blade_raising"]+\
               sawing_times["profile_fixing"])
          
        self.conveyor_speed = 20e3 # mm/min
        self.blade_cost = 170 # euro
        self.Fs = 1350
        
        self.power = 10 # (kW)
        self.productive_costs["energy"] = self.power*cost_energy_min
                
        self.costs["labour"] = self.nw*wage_per_minute["machine"]
        
        #self.consumables()
        
    def non_productive_time(self,L,bevels=0):
        """ Calculate non-productive time 
            input: bevels .. number of bevels (0,1 or 2)
                    L .. length of the member [mm]
        """
        
        self.times["non-productive"] = sawing_base_time + saw_bevel_time*bevels + L/self.conveyor_speed
        
        return self.times["non-productive"]    
        
    def productive_time(self,steel_grade,h,t,Ah):
        """ Productive time of sawing
            input: h .. vertical part of the material, where thickness of sawed
                        material is comparable to plate thickness [mm]
                   t ..plate thickness corresponding to h
                Ah .. area of the horizontal part of the profile [mm2]
        """
        #self.times["productive"] = 0.0
        Sm = sawing_parameters[steel_grade]["Sm"]
        Q = sawing_parameters[steel_grade]["Q"]
        # calculate productive time for obth ends
        #for i, T in enumerate(t):
        #S = self.feeding_speed(T)
            #TPadd = h[i]/Sm/S + Ah[i]/Q
            #self.times["productive"] += TPadd            
        S = self.feeding_speed(t)
        TP = h/Sm/S + Ah/Q
        return TP
        
        
    def consumables(self,At,t,TP,steel_grade='S355',verb=False):
        """ Cost of sawing consumables 
            Input:
                At .. cross-sectional area of the sawed profile
                t .. wall thickness of the sawed profile
        """
        
        St = self.blade_durability(t,steel_grade)
        
        if verb:
            print(f'St = {St:.2f}')
            print(f'At = {At:.2f}')
            print(f'TP = {TP:.2f}')
            print(f'Saw blade = {self.blade_cost:.2f}')
        
        return At/St/TP*self.blade_cost
        
        #return self.consumables_cost
    
    def feeding_speed(self,t):
        """ Speed of the blade
            input: t .. thickness through which the blade travels [mm]
            output: speed [mm/min]
        """
        
        if t <= 5.0:
            S = 120
        elif t <= 10:
            S = 100
        elif t <= 15:
            S = 90
        elif t <= 20:
            S = 80
        elif t <= 25:
            S = 70
        elif t <= 30:
            S = 60
        elif t <= 35:
            S = 50
        else:
            S = 40
        
        return S
    
    def Fsp(self,t):
        """ Sawing parameter Fsp
            input: t .. thickness through which the blade travels [mm]
            output: parameter Fsp
        """
        
        if t <= 5.0:
            Fsp = 0.4
        elif t <= 10:
            Fsp = 0.45
        elif t <= 15:
            Fsp = 0.5
        elif t <= 20:
            Fsp = 0.55
        elif t <= 25:
            Fsp = 0.6
        elif t <= 30:
            Fsp = 0.65
        elif t <= 35:
            Fsp = 0.7
        else:
            Fsp = 0.8
        
        return Fsp
    
    def blade_durability(self,t,steel_grade="S355"):
        """ Durability of the blade 
            t .. thickness of the plate to be sawn
        """
        
        Q = sawing_parameters[steel_grade]["Q"]
        Fsp = self.Fsp(t)
        St = Q*self.Fs*Fsp
        
        return St
    
    def cost(self,L,bevels, steel_grade, H, T, AH, AT):
        """
        

        Parameters
        ----------
        L : float
            Length of member.
        bevels : float/int
            Number of bevels (0, 1 or 2).
        steel_grade : string
            Steel grade, one of the values in sawing_parameters dict.
        h : float
            height of the sawn portion, where the thickness of the plate
            is close to wall thickness.
        T : list
            wall thicknesses of the "thin" portion of the sawn part at each end of the member
        AH : list
            area of the "solid" part of the section at each end.
        AT : list
            total cross_sectional area of the profile at each end.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        
        
        TNP = self.non_productive_time(L,bevels)
        #TP  = self.productive_time(steel_grade, H, T, AH)
        TP = 0.0
        cCS = 0.0
        # calculate productive time and cost of consumables for each end.   
        for t, h, Ah, At in zip(T,H,AH,AT):     
            TPnew = self.productive_time(steel_grade, h, t, Ah)
            cCS += self.consumables(At, t, TPnew, steel_grade)
            TP += TPnew            
                    
        self.times["productive"] = TP
        self.consumables_cost = cCS
        return super().cost()

# Cost center for drilling
class DrillingCost(CostCentre):

    def __init__(self, length=35.0,width=15.0, workers=1.0,
                 eq_invest=520000.0,eq_resale=0.0,eq_maint=1000.0):

        CostCentre.__init__(self, length,width,workers,
                            eq_invest,eq_resale,eq_maint)
        
        self.name = "Drilling"
        
        self.times["non-productive"] = 3.0          
        
        # Labour cost
        self.costs["labour"] = self.nw*wage_per_minute["machine"] # Hourly cost [e/h]
            
    def non_productive_time(self,L=6000,nrows=1,nholes=2):
        """ Calculate non-productive time 
            input: turns .. number of manual turnings of the profile (0,1 or 2)
                    L .. length of the member [mm]
        """
        np_times = {}
        np_times['setup'] = 4 # operator sets up the equipment
        np_times['drill_bit_picking'] = 0.25
        np_times['drill_tip_to_material_surface'] = 0.2
        np_times['move_head_from_hole_to_hole'] = 0.1
        np_times['to_standby_position'] = 0.2
        np_times['release_and_fix'] = 0.2
        
        npt = 0
        for t in np_times.values():
            npt += t
        
        self.times["non-productive"] = L/self.conveyor_speed
        
        return self.times["non-productive"]        


    def ProductiveTime(self,t_p,cut_length):
        """ Calculate the productive time
            input: Lb .. length of steel piece (mm)
            output: productive time (min)
        """
        # Cutting speed according to thickness [mm] of the plate
        if t_p <= 30.0:   # Plasma cutting
            v_cut = (8.9212*t_p**2.0 - 486.87*t_p + 8155.8)     # [mm/min]
        else:   # Flame cutting
            v_cut = -4.1939*t_p + 658.67                        # [mm/min]
            
        TP = cut_length/v_cut                                   # [min]
        self.times["productive"] = TP
        return TP
    
    def Consumables(self,t_p):
        """ Calculate the cost of consumables
            input: t_p .. plate thickness
            
        """
        # Consumables: Gases, plasma electrodes and nozzles
        if t_p <= 30.0:   # Plasma cutting
            # Total consumables cost [e/min]
            cutting_oxygen = 0.071                                       # [m^3/min]
            oxygen_cost = 0.3                                            # [e/min]
            nozzle_unit_cost = 40                                        # [e]
            nozzle_life = 480                                            # [min]
            nozzle_cost = nozzle_unit_cost/nozzle_life                   # [e/min]
            self.productive_costs["consumables"] = cutting_oxygen*oxygen_cost + nozzle_cost
            
        else:   # Flame cutting
            # Total consumables cost [e/min]
            propane_consump = 0.0063                                        # [m^3/min]
            propane_cost = 18.40                                            # [e/m^3]
            preheat_oxygen_consump = 0.025                                  # [m^3/min]
            cutting_oxygen_consump = 1.0e-5*t_p**2.0 + 0.001*t_p + 0.0224   # [m^3/min]
            oxygen_cost = 4.18                                              # [e/m^3]
            self.productive_costs["consumables"] = propane_consump*propane_cost + (preheat_oxygen_consump + cutting_oxygen_consump)*oxygen_cost        
    
    def Energy(self,t_p):
        """ Cost of energy used by the machine """
        
        if t_p <= 30:
            # Plasma cutting
            P_total = 72.0
        else:
            # Flame cutting
            P_total = 0.0
            
        # Cost of energy [e/min]
        self.productive_costs["energy"] = P_total*cost_energy_min

    def cost(self,t_p,cut_length):
        """ Evaluate different costs and run the total cost
            calculation
        """
        self.ProductiveTime(t_p, cut_length)
        self.Consumables(t_p)
        self.Energy(t_p)
        return super().cost()

class Workshop:
    """
        Class for storing workshop data and doing cost calculations
        
        Data to be contained:
            Dict of cost centres.
        
    """
    
    def __init__(self,steel_price={"plates":BASIC_STEEL_PRICE},cost_centres=None):
        """ Constructor """        
    
        self.name = "Steel Workshop"
        
        self.steel_price = steel_price
    
        if cost_centres is None:
            self.cost_centres = {'blasting':BlastingCost(),
                                 'cutting':CuttingCost(),
                                 'assembly_welding':AssemblingCostWeld(),
                                 'sawing':SawingCost(),
                                 'painting':PaintingCost(),
                                 'beam_welding':BeamWeldingCost(),
                                 'drilling':DrillingCost()}
        else:
            self.cost_centres = {}
            for key, value in cost_centres.items():
                self.cost_centres[key] = value
        
    
    def info(self):
        """ Prints a bunch of information regarding the workshop """
        
        for centre in self.cost_centres.values():
            centre.info()


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Sat Sep 25 15:17:02 2021

Class for plane (steel) trusses

@author: kmela
"""
try:
    from metku.frame2d.frame2d import Frame2D, FrameMember, PointLoad, LineLoad, \
        Support, XYHingedSupport, YHingedSupport
    from metku.framefem.elements import EBBeam, EBSemiRigidBeam
    # from fem.elements.eb_semi_rigid_beam import EBSemiRigidBeam
    from metku.sections.steel import HEA

    from metku.framefem import FrameFEM, BeamSection
    from metku.structures.steel.steel_member import SteelMember
    from metku.eurocodes.en1993.en1993_1_8.rhs_joints import RHSKGapJoint, \
        RHSYJoint
except:
    from frame2d.frame2d import Frame2D, FrameMember, PointLoad, LineLoad, \
        Support, XYHingedSupport, YHingedSupport
    from framefem.elements import EBBeam, EBSemiRigidBeam
    # from fem.elements.eb_semi_rigid_beam import EBSemiRigidBeam
    from sections.steel import HEA

    from framefem import FrameFEM, BeamSection
    from structures.steel.steel_member import SteelMember
    from eurocodes.en1993.en1993_1_8.rhs_joints import RHSKGapJoint, RHSYJoint

from loadIDs import LoadIDs

import numpy as np
import matplotlib.pyplot as plt
import math
from matplotlib.patches import Arc

PREC = 3
L_sys = 0.9


class PlaneTruss(Frame2D):

    def __init__(self, origin=(0, 0), span=10000, left_height=1000,
                 mid_height=1500, right_height=1000, bot_chord_dx = 500,
                 num_elements=2,
                 topology='K',
                 ndiv=4,
                 fem_model=None, q=None):
        """ Constructor 
            Parameters:
            ------------
            :param origin: location of the origin (tuple)
            :param num_elements: number of elements per member
            :param fem_model: FrameFEM object
            :param q: load?
    
            :type origin: tuple
            :type num_elements: int
            :type fem_model: FrameFEM (optional)
    
    
            Variables:
            ----------
            :ivar truss:
            :ivar origin:
            :ivar top_chords: list of members constituting the top chord
            :ivar bottom_chords: list of bottom chord members
            :ivar braces: dict of brace members
            :ivar joints: dict of truss joints
            
            :ivar _H0: y-coordinate of the bottom chord
            :ivar _H1: height at
            :ivar _H2: height at
            :ivar _H3: height at
            :ivar L1: span
            :ivar L2: span
            
        
        """
        super().__init__(num_elements=num_elements, fem_model=fem_model)
        
        self.origin = origin
        self.top_chords = []
        self.bottom_chords = []
        self.braces = {}
        self.joints = {}
               
        self.span = span
        
        # Top and bottom chord nodes
        self.top_coords = []
        self.bottom_coords = []
        
        if mid_height < left_height:
            raise ValueError('Truss height at mid-span cannot be smaller than height at left support.')
            
        if mid_height < right_height:
            raise ValueError('Truss height at mid-span cannot be smaller than height at right support.')
        
        self._H0 = left_height
        self._H1 = mid_height
        self._H2 = right_height
        
        self.n = 1
        self._dx = bot_chord_dx
        
        self.top = topology
        self.ndiv = ndiv
        
        self.generate_topology(topology,ndiv)
        self.generate_joints()
        #self.generate_chords()
        #self.generate_webs()

        #if q is not None:
        #    for tc in self.top_chords:
        #        self.add(LineLoad(tc, [q, q], 'y'))

    @property
    def H0(self):
        return self._H0

    @property
    def H1(self):
        return self._H1

    @H1.setter
    def H1(self, val):
        """
        Sets top chord's height of the leftmost eave
        :param val: distance between top and bottom chord
        """
        
        if len(self.top_chords) > 0:
            tc = self.top_chords[0]
            locs = tc.locs
            tc.n1.coord = np.asarray(self.origin) + np.array([0, self.H0 + val])
            for loc, node in zip(locs, tc.nodes.values()):
                node.coord = tc.to_global(loc)
                for j in tc.joints:
                    j.calc_nodal_coordinates()

        if self.H1 == self.H3:
            self.H3 = val
        self._H1 = val

    @property
    def H2(self):
        return self._H2

    @H2.setter
    def H2(self, val):
        """
        Sets top chord's height on ridge
        :param val: distance between top and bottom chord
        """
        if len(self.top_chords) > 0:
            tc = self.top_chords[0]
            tc1 = self.top_chords[1]
            locs = tc.locs
            locs1 = tc1.locs
            # Set the ridge's height
            tc.n2.coord = np.asarray(self.origin) + np.array(
                [self.L1, self.H0 + val])
    
            # Left side
            for loc, node in zip(locs, tc.nodes.values()):
                node.coord = tc.to_global(loc)
            for j in tc.joints:
                j.calc_nodal_coordinates()
    
            # Right side
            for loc, node in zip(locs1, tc1.nodes.values()):
                node.coord = tc1.to_global(loc)
            for j in tc1.joints:
                j.calc_nodal_coordinates()

        self._H2 = val

    @property
    def H3(self):
        return self._H3

    @H3.setter
    def H3(self, val):
        """
        Sets top chord's height of the rightmost eave
        :param val: distance between top and bottom chord
        """
        if len(self.top_chords) > 0:
            tc = self.top_chords[1]
            locs = tc.locs
            tc.n2.coord = np.asarray(self.origin) + np.array(
                [self.L1 + self.L2, self.H0 + val])
            for loc, node in zip(locs, tc.nodes.values()):
                node.coord = tc.to_global(loc)
            for j in tc.joints:
                j.calc_nodal_coordinates()
        self._H3 = val

    @property
    def dx(self):
        return self._dx

    @dx.setter
    def dx(self, val):

        if len(self.bottom_chords) > 0:
            bc = self.bottom_chords[0]
            locs = bc.locs
            bc.n1.coord = np.asarray(self.origin) + np.array([val, self.H0])
            bc.n2.coord = np.asarray(self.origin) + np.array(
                [self.L1 + self.L2 - val, self.H0])
            for loc, node in zip(locs, bc.nodes.values()):
                node.coord = bc.to_global(loc)
            for j in bc.joints:
                j.calc_nodal_coordinates()
        self._dx = val

    @property
    def top_chord_angle(self):
        """ Angle of the top chord with respect to horizontal axis """
        
        a = (self.H1-self.H0)/(0.5*self.span)
        
        return np.arctan(a)
    
    def top_chord_direction_vector(self,half='left'):
        """ Unit vector in the direction of the top chord """
        
        a = self.top_chord_angle
        
        if half == 'left':
            v = np.array([np.cos(a),np.sin(a)])
        else:
            v = np.array([np.cos(a),-np.sin(a)])
        
        return v

    def generate_topology(self,topology='K',ndiv=4):
        """ Generates truss members for a given topology 
        
            :param: topology .. 'K', 'N', or 'KT'
            :param: ndiv .. number of divisions on one half of the truss
            
        """
        
        if topology == 'K':
            # Generate nodes first
            
            # Top chord nodes:
            S = np.linspace(0,1,ndiv+1)
                        
            X0 = np.array([self.origin[0],self.origin[0]+self.H0])
            X1 = np.array([self.origin[0]+0.5*self.span,self.origin[0]+self.H1])            
            X2 = np.array([self.origin[0]+self.span,self.origin[0]+self.H0])
            
            Xtop = [X0+s*(X1-X0) for s in S]
            Xtop += [X1+s*(X2-X1) for s in S[1:]]
            
            # Bottom chord nodes:
            S = np.linspace(0,1,ndiv)
            dx = 0.5*(Xtop[1][0]-Xtop[0][0])            
            X0 = np.array([self.origin[0]+dx,self.origin[0]])
            X1 = np.array([self.origin[0]+0.5*self.span-dx,self.origin[0]])  
            X2 = np.array([self.origin[0]+self.span-dx,self.origin[0]])
            
            Xbot = [X0+s*(X1-X0) for s in S]
            X1 = np.array([self.origin[0]+0.5*self.span+dx,self.origin[0]])  
            Xbot += [X1+s*(X2-X1) for s in S]
        
                
        
            # Generate top chord members
            for i in range(len(Xtop)-1):                
                self.add(TopChord([Xtop[i],Xtop[i+1]]))
        
            # Generate bottom chord members
            for i in range(len(Xbot)-1):
                self.add(BottomChord([Xbot[i],Xbot[i+1]]))
        
            # Generate braces
            
            # First half
            for i in range(ndiv):
                self.add(TrussBrace([Xtop[i],Xbot[i]]))
                self.add(TrussBrace([Xbot[i],Xtop[i+1]]))
            
            # Second half            
            for i in range(ndiv):                
                self.add(TrussBrace([Xtop[ndiv+i],Xbot[ndiv+i]]))
                self.add(TrussBrace([Xbot[ndiv+i],Xtop[ndiv+i+1]]))
            
            
        """    
        fig, ax = plt.subplots(1)
        
        ax.plot([X[0] for X in Xtop], [X[1] for X in Xtop],'ob')
        ax.plot([X[0] for X in Xbot], [X[1] for X in Xbot],'ob')
        ax.set_aspect('equal')
        """  
        self.top_coords = Xtop
        self.bottom_coords = Xbot    
            
    def generate_joints(self):
        """ Generates joints for the truss """

        # Top chord joints        
        for node in self.top_coords:            
            joint_braces = []
            for brace in self.braces.values():            
                if list(node) in brace.coordinates:
                    joint_braces.append(brace)
            
            joint_chord = []
            for chord in self.top_chords:                                
                if all(np.isin(node,chord.coordinates)):                    
                    joint_chord.append(chord)
            
            self.add(TrussJoint(node,joint_chord,joint_braces))
        
        # Bottom chord joints        
        for node in self.bottom_coords:            
            joint_braces = []
            for brace in self.braces.values():            
                if list(node) in brace.coordinates:
                    joint_braces.append(brace)
            
            joint_chord = []
            for chord in self.bottom_chords:                                
                if all(np.isin(node,chord.coordinates)):                    
                    joint_chord.append(chord)
            
            self.add(TrussJoint(node,joint_chord,joint_braces))

    def add(self, this):
        """
        Adds object to truss
        :param this: object to be added
        :return:
        """
        # TOP CHORD
        if isinstance(this, TopChord):
            self.top_chords.append(this)
            this.mem_id = len(self.members)
            self.members["TC" + str(this.mem_id)] = this
            """ Add nodal coordinates for the member """
            this.calc_nodal_coordinates(self.num_elements)

        # BOTTOM CHORD   
        elif isinstance(this, BottomChord):
            self.bottom_chords.append(this)
            this.mem_id = len(self.members)
            self.members["BC" + str(this.mem_id)] = this
            """ Add nodal coordinates for the member """
            this.calc_nodal_coordinates(self.num_elements)

        # BRACE
        elif isinstance(this, TrussBrace):            
            
            # Check if there's already a joint at given location
            # if not, create a new joint
            
            """
            j1, j2 = None, None
            R = 0.03
            for joint in bottom_chord.joints:
                if np.sqrt((joint.loc - this.bot_loc) ** 2) <= R:
                    j1 = joint
                    this.j1 = j1
                    break
            else:
                j1 = TrussJoint(bottom_chord, this.bot_loc)
                bottom_chord.joints.append(j1)
                this.j1 = j1

            for joint in top_chord.joints:
                if np.sqrt((joint.loc - this.top_loc) ** 2) <= R:
                    j2 = joint
                    this.j2 = j2
                    break
            else:
                j2 = TrussJoint(top_chord, this.top_loc)
                top_chord.joints.append(j2)
                this.j2 = j2
            
            """
            
            # Assign member id
            this.mem_id = len(self.braces)
            # Assign joint id's   
            """
            if j1.jid == None:
                j1.jid = len(self.joints)
                self.joints[j1.jid] = j1
            if j2.jid == None:
                j2.jid = len(self.joints)
                self.joints[j2.jid] = j2            

            j1.add_web(this, self.num_elements)
            j2.add_web(this, self.num_elements)
            """
            self.braces[this.mem_id] = this
            self.members["B" + str(this.mem_id)] = this

        elif isinstance(this,TrussJoint):
            
            this.jid = len(self.joints)
            self.joints[this.jid] = this

        else:
            super().add(this)

    
    def calculate(self, load_id=LoadIDs.ULS):
        super().calculate(load_id)
        for joint in self.joints.values():
            joint.update_forces()
    
    def generate0(self):
        """ Generates the truss
        """
        if not self.is_generated:
            self.is_generated = True
        # Generate members' nodes and elements

        for tc in self.top_chords:
            tc.calc_nodal_coordinates()
            tc.round_coordinates()
            tc.generate(self.f)

        for bc in self.bottom_chords:
            bc.calc_nodal_coordinates()
            bc.round_coordinates()
            bc.generate(self.f)

        for w in self.braces.values():
            w.calc_nodal_coordinates()
            w.round_coordinates()
            w.generate(self.f)

        for member in self.members.values():
            # if member.mtype != "web":
            #     member.calc_nodal_coordinates()
            # member.round_coordinates()
            # member.generate(self.f)
            # # Save coordinates to truss' list
            for coord in member.nodal_coordinates:
                if coord not in self.nodal_coordinates:
                    self.nodal_coordinates.append(coord)
            self.nodes.extend(list(member.nodes.values()))

        # Remove duplicate nodes
        self.nodes = list(set(self.nodes))
        # Generate joints' nodes and elements
        for joint in self.joints.values():
            joint.generate(self.f)

        for support in self.supports.values():
            support.add_support(self.f)
            
        # Add point loads (if any)
        for pointLoad in self.point_loads.values():
            pointLoad.add_load(self.f)
            # Creates new loadcase if one with same load_id does not exist
            lcase_ids = [lc.load for lc in self.f.loadcases.values()]
            if pointLoad.load_id not in lcase_ids:
                self.f.add_loadcase(supp_id=1,
                                    load_id=pointLoad.load_id)

        """
        if self_weight:
            self.add_self_weight()
        """
        # Add line loads (if any)
        for lineLoad in self.line_loads.values():
            member = lineLoad.member
            lineLoad.element_ids = member.lineload_elements(
                lineLoad.coordinates)
            lineLoad.add_load(self.f)
            # Creates new loadcase if one with same load_id does not exist
            lcase_ids = [lc.load for lc in self.f.loadcases.values()]
            if lineLoad.load_id not in lcase_ids:
                self.f.add_loadcase(supp_id=1,
                                    load_id=lineLoad.load_id)

        """
        for pointLoad in self.point_loads.values():
            pointLoad.add_load(self.f)

        for lineLoad in self.line_loads.values():
            member = lineLoad.member
            lineLoad.element_ids = member.lineload_elements(
                lineLoad.coordinates)
            lineLoad.add_load(self.f)
        """

    def mirror(self):
        top_coords = self.top_chord.coordinates
        top_coords = [[top_coords[0][0], top_coords[1][1]],
                      [top_coords[1][0], top_coords[0][1]]]

        self.top_chord.coordinates = top_coords
        bottom_coords = self.bottom_chord.coordinates
        bottom_coords = [[bottom_coords[0][0], bottom_coords[1][1]],
                         [bottom_coords[1][0], bottom_coords[0][1]]]
        self.bottom_chord.coordinates = bottom_coords


class TrussMember(FrameMember):
    def __init__(self,
                 coordinates,
                 mem_id="",
                 profile="SHS 100x5",
                 material="S355",
                 num_elements=2,
                 Sj1=np.inf,
                 Sj2=np.inf, **kwargs):

        self.joints = []
        # FrameMember -objects that chord is connected to
        self.columns = {}
        # List of chords' members between joints
        self.steel_members = []
        super().__init__(coordinates,
                         mem_id,
                         profile,
                         material,
                         num_elements,
                         Sj1,
                         Sj2, **kwargs)

    @property
    def profile(self):
        return super().profile

    @profile.setter
    def profile(self, val):
        super(TrussMember, self.__class__).profile.fset(self, val)
        for joint in self.joints:
            joint.calc_nodal_coordinates()

    def calc_nodal_coordinates(self, num_elements=0):
        """
            Calculates node locations along member
            Coordinates are global coordinates
            Adds locations to a list where they can be accessed later
            :param num_elements: int, number of elements to be created
        """
        self.nodal_coordinates.clear()
        self.loc.clear()
        self.nodal_coordinates.extend(self.added_coordinates)
        if num_elements > 0:
            self.num_elements = num_elements
        c00, c01 = self.coordinates
        
        # Create calculation nodes for joints (if joints are present)
        if len(self.joints) > 0:
            joint_locs = [j.loc for j in self.joints]
            if 0 not in joint_locs:
                joint_locs.append(0)
            if 1 not in joint_locs:
                joint_locs.append(1)
            joint_locs.sort()
            for i in range(len(joint_locs) - 1):
                j_loc_0 = joint_locs[i]
                j_loc_1 = joint_locs[i + 1]

                d_loc = j_loc_1 - j_loc_0
                d_loc /= self.num_elements
                for j in range(self.num_elements + 1):
                    loc = j_loc_0 + d_loc * j
                    loc = round(loc, 5)
                    coord = list(self.to_global(loc))
                    loc = min(loc, 1)
                    if loc not in self.loc:
                        self.loc.append(loc)
                    if loc <= 1 and coord not in self.nodal_coordinates:
                        self.nodal_coordinates.append(coord)

            last_joint_coord = list(self.to_global(joint_locs[-1]))
            if last_joint_coord not in self.nodal_coordinates:
                self.nodal_coordinates.append(last_joint_coord)
            if list(self.to_global(1)) not in self.nodal_coordinates:
                self.nodal_coordinates.append(list(self.to_global(1)))
            c0, c1 = self.coordinates
            self.nodal_coordinates.sort(
                key=lambda c: np.linalg.norm(np.asarray(c0) - np.asarray(c)))
            joint_locs.sort()
            coords = [self.to_global(loc) for loc in self.loc if
                      loc not in joint_locs]

        else:
            # If no joints are present, use the method of the FrameMember class
            #print("use FrameMember calc method")
            super().calc_nodal_coordinates(num_elements)

        if self.is_generated:
            nodes = [n for n in self.nodes.values()]
            nodes = sorted(nodes, key=lambda n: np.linalg.norm(
                np.asarray(c00) - np.asarray(n.coord)))
            self.loc.sort()

            for loc, node in zip(self.locs, nodes):
                node.coord = self.to_global(loc)

    def check_cross_section(self):

        for smem in self.steel_members:
            smem.profile = self.cross_section

        if not isinstance(self.TrussBrace): # self.mtype != 'web':  # If member is chord
            # Sort joints using joints' local coordinate
            self.joints.sort(key=lambda x: x.loc)
            if not len(self.steel_members):
                for i in range(len(self.joints) - 1):
                    j0 = self.joints[i]
                    j1 = self.joints[i + 1]
                    x0, y0 = j0.coordinate
                    x1, y1 = j1.coordinate
                    L = np.sqrt((x0 - x1) ** 2 + (y0 - y1) ** 2)
                    steel_mem = SteelMember(self.cross_section, L,
                                            [L_sys, L_sys])
                    for id in range(j0.nodes['c0'].nid, j1.nodes['c0'].nid):
                        N, V, M = self.nodal_forces[id]
                        steel_mem.add_section(ned=N, myed=M, vyed=V)
                    self.steel_members.append(steel_mem)
            else:
                for i in range(len(self.joints) - 1):
                    j0 = self.joints[i]
                    j1 = self.joints[i + 1]
                    x0, y0 = j0.coordinate
                    x1, y1 = j1.coordinate
                    L = np.sqrt((x0 - x1) ** 2 + (y0 - y1) ** 2)
                    steel_mem = self.steel_members[i]
                    steel_mem.length = L
                    steel_mem.clear_sections()
                    for id in range(j0.nodes['c0'].nid,
                                    j1.nodes['c0'].nid):
                        N, V, M = self.nodal_forces[id]
                        steel_mem.add_section(ned=N, myed=M, vyed=V)
                # r = []
                # r_sect = steel_member.check_sections()
                # r_buckling = steel_member.check_buckling()
                # r_lt_buckling = steel_member.check_LT_buckling()
                # r.extend(r_sect)
                # r.extend(r_buckling)
                # r.append(r_lt_buckling)
        else:
            super().check_cross_section()

    def local(self, value):
        """
        Returns chord's local coordinate in global coordinates
        :param value: local coordinate, 0 <= value <= 1
        """
        if 0 >= value >= 1:
            raise ValueError('Value must be between 0 and 1')
        start_node, end_node = self.coordinates

        v = np.array([math.cos(self.angle), math.sin(self.angle)])

        x0, y0 = start_node
        Lx = self.length * value

        return list(np.asarray(start_node) + Lx * v)

    def global_coord(self, value):
        """
        Return's global coordinate in chord's local coordinates
        """
        x, y = value
        x0, y0 = self.coordinates[0]

        L = np.sqrt((x0 - x) ** 2 + (y0 - y) ** 2)
        local = L / self.length

        return local

    def shape(self, x):
        """
        Returns chord's y value corresponding to x
        """
        start_node, end_node = self.coordinates
        x0, y0 = start_node
        Lx = end_node[0] - start_node[0]
        Ly = end_node[1] - start_node[1]
        try:
            k = Ly / Lx
        except ZeroDivisionError:
            k = 0

        return k * (x - x0) + y0

    # @property
    # def angle(self):
    #     """
    #     Returns member's angle in radians
    #     """
    #     start_node, end_node = self.coordinates
    #     x0, y0 = start_node
    #     x1, y1 = end_node
    #     if (x1 - x0) == 0:
    #         angle = math.radians(90)
    #     else:
    #         angle = math.atan((y1 - y0) / (x1 - x0))
    #     return angle


    def plot(self, print_text=True, color=True, axes=None):

        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes        

        X = self.coordinates
        if color:
            if self.is_strong_enough:
                color = 'green'
            else:
                color = 'red'
        else:
            color = 'k'
        # Plot members
        ax.plot([X[0][0], X[1][0]], [X[0][1], X[1][1]], color)
        # Plot text

        # Calculate text location
        delta_y = X[1][1] - X[0][1]
        delta_x = X[1][0] - X[0][0]
        if delta_x == 0:
            rot = 90
        elif delta_y == 0:
            rot = 0
        else:
            rot = math.degrees(math.atan(delta_y / delta_x))
            # rot = self.angle

        x = (X[0][0] + X[1][0]) / 2
        y = (X[1][1] + X[0][1]) / 2
        horzalign = 'center'
        vertalign = 'center'
        if print_text:
            ax.text(x, y, str(self.mem_id) + ": " + str(self.cross_section),
                     rotation=rot, horizontalalignment=horzalign,
                     rotation_mode='anchor')
        else:
            ax.text(x, y, str(self.mem_id),
                     rotation=rot, horizontalalignment=horzalign,
                     verticalalignment=vertalign)

    def bmd0(self, scale):
        """
        Plots bending moment diagram
        :param scale: float, scaling factor
        """
        X = []
        Y = []
        self.calc_nodal_forces()

        moment_values = [x[2] for x in self.nodal_forces.values()]
        node_ids = list(self.nodes.keys())
        x = self.nodal_coordinates[0][0]
        y = self.nodal_coordinates[0][1]
        X.append(x)
        Y.append(y)
        for i in range(len(self.nodes)):
            node = node_ids[i]
            if self.mtype == "top_chord" or self.mtype == "bottom_chord":
                x0 = self.nodal_coordinates[i][0]
                y0 = self.nodal_coordinates[i][1]
                bending_moment = self.nodal_forces[node][2]
                y1 = bending_moment / (1000 / scale)
                x = x0
                y = y0 - y1
                X.append(x)
                Y.append(y)
                if i == 0 or i == len(self.nodes) - 1 or bending_moment == max(
                        moment_values) or \
                        bending_moment == min(moment_values):
                    if bending_moment > 0:
                        vert = 'top'
                    else:
                        vert = 'bottom'
                    plt.text(x, y, f'{bending_moment:.2f} kNm',
                             verticalalignment=vert)

            elif self.mtype == "web":
                x0 = self.nodal_coordinates[i][0]
                y0 = self.nodal_coordinates[i][1]
                bending_moment = self.nodal_forces[node][2]
                x1 = bending_moment / (1000 / scale)
                x = x0 + x1
                y = y0
                X.append(x)
                Y.append(y)
                if i == 0 or i == len(self.nodes) - 1 or bending_moment == max(
                        moment_values) or \
                        bending_moment == min(moment_values):
                    if bending_moment > 0:
                        horz = 'left'
                    else:
                        horz = 'right'
                    plt.text(x, y, f'{bending_moment:.2f} kNm',
                             horizontalalignment=horz)
        X.append(self.nodal_coordinates[i][0])
        Y.append(self.nodal_coordinates[i][1])
        plt.plot(X, Y, color='gray')


class Chord(FrameMember):

    def __init__(self, coordinates, **kwargs):
        super().__init__(coordinates, **kwargs)

        self.members = []
        self.joints = []

    def generate0(self, fem_model):

        locs = [0, 1]
        for joint in self.joints:
            if joint.loc not in locs:
                locs.append(joint.loc)

        locs.sort()
        for i in range(len(locs) - 1):
            c0 = self.to_global(locs[i])
            c1 = self.to_global(locs[i + 1])
            mem = FrameMember([c0, c1])
            self.members.append(mem)
            mem.generate(fem_model)

    def check_cross_section(self):
        for mem in self.members:
            mem.check_cross_section()
    
    def calc_nodal_coordinates(self, num_elements=2):
        """
            Calculates node locations along a chord member
            Coordinates are global coordinates
            Adds locations to a list where they can be accessed later
            :param num_elements: int, number of elements to be created
        """
        self.nodal_coordinates.clear()
        self.loc.clear()
        print(self.added_coordinates)
        self.nodal_coordinates.extend(self.added_coordinates)
        if num_elements:
            self.num_elements = num_elements
        c00, c01 = self.coordinates
        joint_locs = [j.loc for j in self.joints]
        if len(joint_locs):
            if 0 not in joint_locs:
                joint_locs.append(0)
            if 1 not in joint_locs:
                joint_locs.append(1)
            joint_locs.sort()
            for i in range(len(joint_locs) - 1):
                j_loc_0 = joint_locs[i]
                j_loc_1 = joint_locs[i + 1]

                d_loc = j_loc_1 - j_loc_0
                d_loc /= self.num_elements
                for j in range(self.num_elements + 1):
                    loc = j_loc_0 + d_loc * j
                    loc = round(loc, 5)
                    coord = list(self.to_global(loc))
                    loc = min(loc, 1)
                    if loc not in self.loc:
                        self.loc.append(loc)
                    if loc <= 1 and coord not in self.nodal_coordinates:
                        self.nodal_coordinates.append(coord)

            last_joint_coord = list(self.to_global(joint_locs[-1]))
            if last_joint_coord not in self.nodal_coordinates:
                self.nodal_coordinates.append(last_joint_coord)
            if list(self.to_global(1)) not in self.nodal_coordinates:
                self.nodal_coordinates.append(list(self.to_global(1)))
            c0, c1 = self.coordinates
            self.nodal_coordinates.sort(
                key=lambda c: np.linalg.norm(np.asarray(c0) - np.asarray(c)))
            joint_locs.sort()
            coords = [self.to_global(loc) for loc in self.loc if
                      loc not in joint_locs]

        else:
            super().calc_nodal_coordinates(num_elements)

        if self.is_generated:
            nodes = [n for n in self.nodes.values()]
            nodes = sorted(nodes, key=lambda n: np.linalg.norm(
                np.asarray(c00) - np.asarray(n.coord)))
            self.loc.sort()

            for loc, node in zip(self.locs, nodes):
                node.coord = self.to_global(loc)


class TopChord(TrussMember):
    def __init__(self,
                 coordinates,
                 **kwargs):
        super().__init__(coordinates,
                         **kwargs)

        self.mtype = 'top_chord'
        self.steel_members = []


class BottomChord(TrussMember):

    def __init__(self,
                 coordinates,
                 **kwargs):
        super().__init__(coordinates,
                         **kwargs)
        self.mtype = 'bottom_chord'
        self.steel_members = []

    @property
    def perpendicular(self):
        return -1 * super().perpendicular


class TrussBrace(TrussMember):
    """ Class for truss web members, or braces """
    def __init__(self,
                 coordinates,
                 mem_id="",
                 profile="SHS 50x2",
                 material="S355",
                 num_elements=1,
                 Sj1=0,
                 Sj2=0):
        """ Constructor
        
            :param: bot_loc .. location of the member end at the bottom chord
            :param: top_loc .. location of the member end at the top chord
            :param: coord_type .. 'local' (default) or 'global' coordinates
        """

        self.j1 = None
        self.j2 = None
        
        self.top_joint = None
        self.bottom_joint = None

        super().__init__(coordinates, profile=profile,material=material, num_elements=num_elements)
   
        self.top_coord = self.coordinates[1]
        self.bot_coord = self.coordinates[0]
        self.Sj1 = Sj1
        self.Sj2 = Sj2
        self.mtype = 'brace'
        self.steel_members = [self.member] 


class TrussJoint:
    """ Class for welded tubular joints """

    def __init__(self, node, chord, brace, reverse=False):
    
        for b in brace:
            b.joints.append(self)
            
            if isinstance(chord[0],TopChord):
                b.j1 = self
            else:
                b.j2 = self
            
        for c in chord:
            c.joints.append(self)

        self.chord = chord
        self.helper_chord = chord.copy()
        self.chord_elements = {}
        self.jid = None
 
        self.cnode = node
        self.chord_nodes = []
        self.joint_type = None
        self.nodal_coordinates = {}
        self.nodes = {}
        self.elements = {}
        self.element_ids = []
        self.braces = brace
        self.is_generated = False
        # Index for eccentricity elements' material and section
        self.idx = None
        self.rhs_joint = None
        self.r = []
        self.__e = 0
        self.g_set = False
        self.reverse = reverse

    def add_node_coord(self, coord):
        coord = [round(c, PREC) for c in coord]
        if coord not in self.nodal_coordinates:
            self.nodal_coordinates.append(coord)

    def add_brace(self, brace):
        self.braces[brace.mem_id] = brace

        # if self.is_generated:
        self.calc_nodal_coordinates()

    @property
    def loc(self):
        return round(self.__loc, PREC)

    @loc.setter
    def loc(self, val):
        # if self.chord.mtype == 'bottom_chord' or 0.01 <= self.loc <= 0.99:
        min_val = self.g1 / self.chord.length
        max_val = 1 - self.g1 / self.chord.length

        min_val = 0
        max_val = 1

        val = min(max(min_val, val), max_val)
        val = round(val, PREC)

        # locs = [j.loc for j in self.chord.joints]
        # idx = locs.index(self.loc)
        # if idx == 0:
        #     min_val = 1e-2
        # else:
        #     min_val = locs[idx - 1] + 0.05
        # if idx == len(locs) -1:
        #     max_val = 1-1e-2
        # else:
        #     max_val = locs[idx + 1] - 0.05
        #
        # val = max(min(val, max_val), min_val)

        if self.reverse:
            val = 1 - val

        if val != self.loc:
            self.__loc = val
            # if self.chord.mtype == 'top_chord':
            #     for web in self.webs.values():
            #         web.top_loc = self.loc
            # else:
            #     for web in self.webs.values():
            #         web.bot_loc = self.loc

            self.calc_nodal_coordinates()

    @property
    def coordinate(self):
        if self.chord.mtype == 'bottom_chord':
            return [round(c, PREC) for c in
                    self.helper_chord.to_global(self.loc)]

        return [round(c, PREC) for c in self.chord.to_global(self.loc)]

    @coordinate.setter
    def coordinate(self, val):
        # (x, y) -coordinate
        if isinstance(val, list):
            local = self.chord.global_coord(val)
            self.coordinate = local
        # local coordinate
        elif isinstance(val, float):
            if 0 <= val <= 1:
                self.loc = val
                self.calc_nodal_coordinates()

    @property
    def theta1(self):
        if self.chord.mtype == 'top_chord' and not self.chord.reverse:
            webs = sorted(self.webs.values(), key=lambda w: w.bot_loc)
        else:
            webs = sorted(self.webs.values(), key=lambda w: w.top_loc)
        w1 = webs[0]
        theta1 = min(abs(w1.angle - self.chord.angle),
                     abs(np.pi - w1.angle - self.chord.angle))

        return theta1

    @property
    def theta2(self):
        if len(self.webs) >= 2:
            if self.chord.mtype == 'top_chord' and not self.chord.reverse:
                webs = sorted(self.webs.values(), key=lambda w: w.bot_loc)
            else:
                webs = sorted(self.webs.values(), key=lambda w: w.top_loc)
            w2 = webs[1]
            theta2 = min(abs(w2.angle - self.chord.angle),
                         abs(np.pi - w2.angle - self.chord.angle))
            return theta2
        return 0

    @property
    def e(self):
        """ Eccentrictity of the connection """

        # if self.joint_type == 'K':
        #     w1, w2 = self.webs.values()
        #     ec = w1.line_intersection(w2)
        #     # TODO: distance from ec to chord's axis

        if self.g_set:
            if self.joint_type == 'K':
                w1, w2 = list(self.webs.values())
                h1 = w1.h
                h2 = w2.h
                h0 = self.chord.h
                theta1 = min(abs(w1.angle - self.chord.angle),
                             abs(np.pi - w1.angle - self.chord.angle))
                theta2 = min(abs(w2.angle - self.chord.angle),
                             abs(np.pi - w2.angle - self.chord.angle))
                e = h1 / (2 * np.sin(theta1)) + h2 / (
                        2 * np.sin(theta2)) + self.__g1
                e *= (np.sin(theta1) * np.sin(theta2)) / np.sin(
                    theta1 + theta2)
                e -= h0 / 2

                return e

            return 0

        else:
            return self.__e

    @e.setter
    def e(self, val):
        self.g_set = False
        delta = val - self.e
        # self.calc_nodal_coordinates()
        if self.joint_type == 'K':
            w1, w2 = self.webs.values()
            h1 = w1.h
            h2 = w2.h
            h0 = self.chord.h

            X1 = np.cos(self.theta1) * w1.length
            X2 = np.cos(self.theta2) * w2.length

            theta1 = np.arctan(self.theta1 + delta / X1)
            theta2 = np.arctan(self.theta2 + delta / X2)

            # theta1 = min(abs(w1.angle - self.chord.angle),
            #              abs(np.pi - w1.angle - self.chord.angle))
            # theta2 = min(abs(w2.angle - self.chord.angle),
            #              abs(np.pi - w2.angle - self.chord.angle))

            g = val + h0 / 2
            g *= np.sin(theta1 + theta2) / (np.sin(theta1) * np.sin(theta2))
            g -= h1 / (2 * np.sin(theta1))
            g -= h2 / (2 * np.sin(theta2))
            self.__g1 = g

            if self.chord.mtype == 'top_chord':
                w1.j1.calc_nodal_coordinates()
                w2.j1.calc_nodal_coordinates()
            else:
                w1.j2.calc_nodal_coordinates()
                w2.j2.calc_nodal_coordinates()

        self.__e = val

    @property
    def g1(self):

        if self.joint_type == 'K':
            w1, w2 = self.webs.values()
            h1 = w1.h
            h2 = w2.h
            h0 = self.chord.h

            theta1 = min(abs(w1.angle - self.chord.angle),
                         abs(np.pi - w1.angle - self.chord.angle))
            theta2 = min(abs(w2.angle - self.chord.angle),
                         abs(np.pi - w2.angle - self.chord.angle))

            g = self.e + h0 / 2
            g *= np.sin(theta1 + theta2) / (np.sin(theta1) * np.sin(theta2))
            g -= h1 / (2 * np.sin(theta1))
            g -= h2 / (2 * np.sin(theta2))
            return g

        else:
            return self.__g1

    @g1.setter
    def g1(self, val):
        self.g_set = True
        self.__g1 = val
        # if self.joint_type == 'K':
        #     w1, w2 = list(self.webs.values())
        #     h1 = w1.h
        #     h2 = w2.h
        #     h0 = self.chord.h
        #
        #     e = h1 / (2 * np.sin(self.theta1)) + h2 / (2 * np.sin(self.theta2)) + self.__g1
        #     e *= (np.sin(self.theta1) * np.sin(self.theta2)) / np.sin(self.theta1 + self.theta2)
        #     e -= h0 / 2
        #
        #     self.e = e

        self.calc_nodal_coordinates()

    @property
    def g2(self):
        return self.__g2

    @g2.setter
    def g2(self, val):
        self.__g2 = val
        self.calc_nodal_coordinates()

    @property
    def N0(self):
        """
        Returns maximum normal force on chord on gap

        :return: max N
        """
        if self.joint_type == 'Y':
            N0 = self.chord.nodal_forces[self.nodes['c0'].nid][0]


        elif self.joint_type == 'K':
            N01 = self.chord.nodal_forces[self.nodes['c01'].nid][0]
            N02 = self.chord.nodal_forces[self.nodes['c02'].nid][0]
            N0 = max(abs(N01), abs(N02))
            # N0 *= np.sign(N01)
        else:
            # TODO: KT
            N0 = 0
        return N0
        # N = [abs(elem.shear_force[0]) for elem in self.elements.values()]
        # return max(N)

    @property
    def V0(self):
        """
        Returns maximum shear force on chord on gap

        :return: max N
        """
        V0 = self.chord.nodal_forces[self.nodes['c0'].nid][1]
        return abs(V0)

        # V = [abs(elem.axial_force[0]) for elem in self.elements.values()]
        # return max(V)

    @property
    def M0(self):
        if self.joint_type == 'Y':
            M0 = self.chord.nodal_forces[self.nodes['c0'].nid][2]

        elif self.joint_type == 'K':
            M01 = self.chord.nodal_forces[self.nodes['c01'].nid][2]
            M02 = self.chord.nodal_forces[self.nodes['c02'].nid][2]
            M0 = max(abs(M01), abs(M02))
        else:
            # TODO: KT
            pass
        return abs(M0)

    def calculate(self):
        """
        """
        NEd0 = [elem.axial_force for elem in self.chord.elements.values() if
                self.cnode in elem.nodes]
        # item for sublist in list for item in sublist
        NEd0 = [N for forces in NEd0 for N in forces]
        maxN = max(NEd0)
        minN = min(NEd0)
        if abs(minN) > maxN:
            NEd0 = minN
        else:
            NEd0 = maxN

        if self.joint_type == "K" or self.joint_type == 'N':
            NEd1, NEd2 = [w.ned for w in self.webs.values()]
            wNEd = np.array([NEd1, NEd2])
            # Resistances
            r1 = self.chord_face_failure()
            r2 = self.punching_shear()
            r3 = self.brace_failure()
            r4 = self.chord_shear()
            # Stress ratios
            r1 = abs(wNEd / r1)
            r2 = abs(wNEd / r2)
            r3 = abs(wNEd / r3)
            r5 = abs(NEd0 / r4[1])  # chord
            r4 = abs(wNEd / r4[0])

            # self.r = [r1, r2, r3, r4]
            maxR_for_webs = np.max((r1, r2, r3, r4), axis=0)
            maxR = np.append(maxR_for_webs, r5)
            self.r = maxR
            # return maxR

        elif self.joint_type == 'Y':
            NEd1 = [w.ned for w in self.webs.values()][0]

            r1 = self.chord_face_failure()
            r2 = self.chord_web_buckling()
            r3 = self.brace_failure()

            r1 = abs(NEd0 / r1)
            r2 = abs(NEd0 / r2)
            r3 = abs(NEd1 / r3)
            # self.r = [r1, r2, r3]

            maxR = [r3, max(r1, r2)]
            self.r = maxR
            # return maxR
        elif self.joint_type == 'KT':
            pass

    def chord_face_failure(self):
        if self.rhs_joint:
            results = self.rhs_joint.chord_face_failure()
            return results / 1000

    def punching_shear(self):
        if self.rhs_joint and self.joint_type == 'K':
            results = self.rhs_joint.punching_shear()
            return results / 1000

    def brace_failure(self):
        if self.rhs_joint:
            results = self.rhs_joint.brace_failure()
            return results / 1000  # N to kN

    def chord_shear(self):
        if self.rhs_joint and self.joint_type == 'K':
            V0 = [elem.shear_force for elem in self.chord.elements.values() if
                  self.cnode in elem.nodes]
            # item for sublist in list for item in sublist
            V0 = [abs(V) for forces in V0 for V in forces]
            V0 = max(V0)
            self.rhs_joint.V0 = V0 * 1e3  # kN to N
            results = [r / 1000 for r in self.rhs_joint.chord_shear()]
            return results

    def chord_web_buckling(self):
        if self.rhs_joint and self.joint_type == 'Y':
            return self.rhs_joint.chord_web_buckling() / 1000

    def calc_nodal_coordinates(self):
        """
        Calculates nodal coordinates

        Y:
        -----
        TODO

        K :
        -----
        c0 -- joint's center coordinate
        ec -- coordinate of web's axis' intersection
        w1 -- leftmost web
        c01 -- web's end coordinate's projection to chord's axis
        c1 -- web's end coordinate on chord's flange
        w2 -- rightmost web
        c02 -- web's end coordinate's projection to chord's axis
        c2 -- web's end coordinate on chord's flange

        KT:
        -----
        TODO

        :return:
        """
        coords = []

        # Remove coordinates from chord
        for coord in self.nodal_coordinates.values():
            if list(coord) in self.chord.added_coordinates:
                self.chord.added_coordinates.remove(list(coord))

        if list(self.coordinate) in self.chord.added_coordinates:
            self.chord.added_coordinates.remove(list(self.coordinate))

        # Y
        if len(self.webs) == 1:
            if self.chord.mtype == "top_chord":
                if self.loc == 0:
                    self.loc = self.g1 / self.chord.length
                if self.loc == 1:
                    self.loc = 1 - self.g1 / self.chord.length
            self.joint_type = "Y"
            w1 = list(self.webs.values())[0]
            c0 = self.coordinate
            c1 = c0 + self.chord.h / 2 * self.chord.perpendicular
            if self.chord.mtype == 'top_chord':
                w1.top_coord = c1
            else:
                w1.bot_coord = c1
            coords = [c0, c1]

            self.nodal_coordinates['c0'] = c0
            self.nodal_coordinates['c1'] = c1

            theta1 = min(abs(w1.angle - self.chord.angle),
                         abs(np.pi - w1.angle - self.chord.angle))
            # Set design class
            self.rhs_joint = RHSYJoint(self.chord.cross_section,
                                       w1.cross_section,
                                       np.degrees(theta1))

        # K
        elif len(self.webs) == 2:
            # Set joint type
            self.joint_type = 'K'
            # Center coordinate
            c0 = self.coordinate
            # Eccentricity coordinate
            ec = c0 - self.e * self.chord.perpendicular
            # Calculate webs' and chord's axis' intersections
            if self.chord.mtype == 'top_chord' and not self.chord.reverse:
                w1, w2 = sorted(self.webs.values(), key=lambda w: w.bot_loc)
                c01 = self.chord.line_intersection([w1.bot_coord, ec])
                c02 = self.chord.line_intersection([w2.bot_coord, ec])
            else:
                w1, w2 = sorted(self.webs.values(), key=lambda w: w.top_loc)
                c01 = self.chord.line_intersection([w1.top_coord, ec])
                c02 = self.chord.line_intersection([w2.top_coord, ec])
            # Calculate distance between axis' intersection
            # and projection from web to chord's axis
            theta1 = min(abs(w1.angle - self.chord.angle),
                         abs(np.pi - w1.angle - self.chord.angle))
            theta2 = min(abs(w2.angle - self.chord.angle),
                         abs(np.pi - w2.angle - self.chord.angle))

            d1 = abs(self.chord.h / 2 / np.tan(theta1))
            c01 -= self.chord.unit * d1

            d2 = abs(self.chord.h / 2 / np.tan(theta2))
            c02 += self.chord.unit * d2

            if np.linalg.norm(
                    self.chord.to_global(1) - c01) > self.chord.length:
                d01 = np.linalg.norm(self.chord.to_global(0) - c01)
                c01 += self.chord.unit * d01
                c0 += self.chord.unit * d01
                c02 += self.chord.unit * d01

            elif np.linalg.norm(
                    self.chord.to_global(0) - c02) > self.chord.length:
                d02 = np.linalg.norm(self.chord.to_global(1) - c02)
                c01 -= self.chord.unit * d02
                c0 -= self.chord.unit * d02
                c02 -= self.chord.unit * d02

            if self.chord.mtype == 'top_chord':
                c1 = np.asarray(
                    c01) + self.chord.h / 2 * self.chord.perpendicular
                c2 = np.asarray(
                    c02) + self.chord.h / 2 * self.chord.perpendicular
                w1.top_coord = c1
                w2.top_coord = c2
            else:
                c1 = np.asarray(
                    c01) + self.chord.h / 2 * self.chord.perpendicular
                c2 = np.asarray(
                    c02) + self.chord.h / 2 * self.chord.perpendicular
                w1.bot_coord = c1
                w2.bot_coord = c2

            coords = [c0, c01, c02, c1, c2]

            self.nodal_coordinates['c0'] = c0
            self.nodal_coordinates['c01'] = c01
            self.nodal_coordinates['c02'] = c02
            self.nodal_coordinates['c1'] = c1
            self.nodal_coordinates['c2'] = c2

            # Set design class
            self.rhs_joint = RHSKGapJoint(self.chord.cross_section,
                                          [w1.cross_section,
                                           w2.cross_section],
                                          [np.degrees(theta1),
                                           np.degrees(theta2)],
                                          self.g1)

        # KT
        elif len(self.webs) == 3:
            self.joint_type = 'KT'
            c0 = self.chord.to_global(self.loc)
            ec = c0 - self.e * self.chord.perpendicular
            if self.chord.mtype == 'top_chord':
                w1, w2, w3 = sorted(self.webs.values(),
                                    key=lambda w: w.bot_loc)
            else:
                w1, w2, w3 = sorted(self.webs.values(),
                                    key=lambda w: w.top_loc)
            c01 = self.chord.line_intersection(w1)
            theta1 = w1.angle - self.chord.angle
            d1 = abs(self.chord.h / 2 / np.tan(theta1))
            c01 -= self.chord.unit * d1

            c02 = self.chord.line_intersection(w2)
            theta2 = w2.angle - self.chord.angle
            d2 = abs(self.chord.h / 2 / np.tan(theta2))
            c02 += self.chord.unit * d2

            # TODO

        # Move existing nodes
        if self.is_generated:
            for key, coord in self.nodal_coordinates.items():
                self.nodes[key].coord = coord
        # Add nodes to chord
        else:
            for c in coords:
                self.chord.add_node_coord(c)

    def round_coordinates(self, prec=PREC):
        for name, coord in self.nodal_coordinates.items():
            self.nodal_coordinates[name] = [round(c, prec) for c in coord]

    def generate(self, fem_model):
        """
        """
        self.round_coordinates()
        self.generate_nodes(fem_model)
        self.add_section_and_material(fem_model)
        self.generate_eccentricity_elements(fem_model)
        self.get_chord_elements()

    def generate_nodes(self, fem_model):
        """
        Generates eccentricity nodes to chord
        :param fem_model: FrameFEM -instance, model where nodes are added
        """
        if not self.is_generated:
            self.is_generated = True
            for name, coord in self.nodal_coordinates.items():
                idx = fem_model.nodal_coords.index(coord)
                self.nodes[name] = fem_model.nodes[idx]
                if coord == self.coordinate:
                    self.cnode = fem_model.nodes[idx]
                if coord in self.chord.nodal_coordinates:
                    self.chord_nodes.append(fem_model.nodes[idx])
        # sort chord nodes from leftmost to rightmost
        self.chord_nodes = sorted(self.chord_nodes, key=lambda node: node.x)

    def get_chord_elements(self):
        """
        Gets the joint's elements which are on chord
        """
        if len(self.chord_nodes) > 2:
            for i in range(2):
                nodes = self.chord_nodes[i:i + 2]
                for key in self.chord.elements.keys():
                    element = self.chord.elements[key]
                    if nodes[0] in element.nodes and nodes[1] in element.nodes:
                        self.chord_elements[key] = element

    def add_section_and_material(self, fem_model):
        """
        Add's eccentricity elements section and material properties
        to the fem model.
        Eccentricity elements are modelled as infinitely rigid elements
        """
        self.idx = len(fem_model.sections)

        # Material(E, nu, rho)
        fem_model.add_material(210e3, 0.3, 7850e-9)
        # BeamSection(A, Iy) [m^2, m^4]
        sect = BeamSection(38880 * 1e-6, 6299657109 * 1e-12)
        fem_model.add_section(sect)

        # USE CHORD'S PROPERTIES
        # self.idx = self.chord.mem_id

    def generate_eccentricity_elements(self, fem_model):
        """
        Generates eccentricity elements
        """

        section = HEA(1000)

        index = len(fem_model.elements)
        if self.joint_type == 'Y':
            self.elements[index] = EBBeam(self.nodes['c0'],
                                          self.nodes['c1'],
                                          section,
                                          fem_model.materials[self.idx])
            # Add it to fem model
            fem_model.add_element(self.elements[index])
            self.element_ids.append(index)


        elif self.joint_type == 'K':
            # Chrod and web
            self.elements[index] = EBBeam(self.nodes['c01'],
                                          self.nodes['c1'],
                                          section,
                                          fem_model.materials[self.idx])
            # Add it to fem model
            fem_model.add_element(self.elements[index])
            self.element_ids.append(index)
            # Chrod and web
            self.elements[index + 1] = EBBeam(self.nodes['c02'],
                                              self.nodes['c2'],
                                              section,
                                              fem_model.materials[self.idx])
            # Add it to fem model
            fem_model.add_element(self.elements[index + 1])
            self.element_ids.append(index + 1)

        # TODO: KT
        # elif self.joint_type == 'KT':
        #     cn1, cn2, cn3 = chord_nodes
        #     wn1, wn2, wn3 = web_nodes
        #     # Chrod and web
        #     self.elements[index] = EBBeam(cn1,
        #                                 wn1,
        #                                 section,
        #                                 fem_model.materials[self.idx])
        #     # Add it to fem model
        #     fem_model.add_element(self.elements[index])
        #     self.element_ids.append(index)
        #     # Central and web
        #     self.elements[index+1] = EBBeam(cn2,
        #                                     wn2,
        #                                     section,
        #                                     fem_model.materials[self.idx])
        #     # Add it to fem model
        #     fem_model.add_element(self.elements[index+1])
        #     self.element_ids.append(index+1)
        #     # Chrod and web
        #     self.elements[index+2] = EBBeam(cn3,
        #                                     wn3,
        #                                     section,
        #                                     fem_model.materials[self.idx])
        #     # Add it to fem model
        #     fem_model.add_element(self.elements[index+2])
        #     self.element_ids.append(index+2)

    def plot(self, print_text=True, color=True):

        x, y = self.coordinate
        if color:
            if self.is_strong_enough:
                color = 'green'
            else:
                color = 'red'
        else:
            color = 'cyan'

        horzalign = 'center'
        vertalign = 'center'
        plt.scatter(x, y, s=100, c=color)
        plt.text(x, y, str(self.jid), horizontalalignment=horzalign,
                 verticalalignment=vertalign, fontsize=12)

    def plot_joint(self, length=0.02, show=True):
        """
        Plots joint
        :param length: local coordinate, defines how much members are show
        :param show: boolean, whether to plot or not
        """
        X, Y = self.coordinate
        end_line = None
        col = None
        # CHORD
        x, y = self.chord.to_global(self.loc)
        h = self.chord.h
        C0 = self.chord.to_global(max(self.loc - length, 0))
        C01 = self.chord.to_global(min(self.loc + length, 1))

        d = np.linalg.norm(C01 - C0)

        C1 = C0 + h / 2 * self.chord.perpendicular
        C2 = C0 - h / 2 * self.chord.perpendicular

        C11 = C1 + d * self.chord.unit
        C21 = C2 + d * self.chord.unit

        plt.plot([C0[0], C01[0]], [C0[1], C01[1]], '--', color='lightblue')
        plt.plot([C1[0], C11[0]], [C1[1], C11[1]], 'k')
        plt.plot([C2[0], C21[0]], [C2[1], C21[1]], 'k')

        # WEBS
        for w in self.webs.values():
            d = abs(d)
            h = w.h
            if self.chord.mtype == 'top_chord':
                W0 = w.top_coord
            else:
                W0 = w.bot_coord

            # if np.degrees(w.angle) >= 0:
            #     d += 500
            # else:
            #     d -= 500

            W1 = W0 + h / 2 * w.perpendicular
            W2 = W0 - h / 2 * w.perpendicular
            if self.chord.mtype == 'top_chord':
                W01 = W0 - d * w.unit
                W11 = W1 - d * w.unit
                W21 = W2 - d * w.unit
                W1 = line_intersection([C1, C11], [W1, W11])
                W2 = line_intersection([C1, C11], [W2, W21])
            else:
                W01 = W0 + d * w.unit
                W11 = W1 + d * w.unit
                W21 = W2 + d * w.unit
                W1 = line_intersection([C1, C11], [W1, W11])
                W2 = line_intersection([C1, C11], [W2, W21])

            plt.plot([W0[0], W01[0]], [W0[1], W01[1]], '--', color='lightblue')
            plt.plot([W1[0], W11[0]], [W1[1], W11[1]], 'k')
            plt.plot([W2[0], W21[0]], [W2[1], W21[1]], 'k')

        if self.joint_type == 'K':
            if self.chord.mtype == 'top_chord' and not self.chord.reverse:
                w1, w2 = sorted(self.webs.values(), key=lambda w: w.bot_loc)
            else:
                w1, w2 = sorted(self.webs.values(), key=lambda w: w.top_loc)

            theta0 = np.degrees(self.chord.angle)
            theta1 = np.degrees(self.theta1)
            theta2 = np.degrees(self.theta2)

            if self.chord.mtype == "bottom_chord":
                xy1 = w1.bot_coord  # - w1.cross_section.H / 2 * self.chord.unit
                xy2 = w2.bot_coord  # + w2.cross_section.H / 2 * self.chord.unit
                arc1 = Arc(xy1, 100, 100, theta1=theta0 + 180 - theta1,
                           theta2=theta0 + 180)
                arc2 = Arc(xy2, 100, 100, theta1=theta0,
                           theta2=theta0 + theta2)

                x1, y1 = (xy1 + np.array([-50, -10])) + 50 * w1.unit
                plt.text(x1, y1, f"{theta1:.1f}\N{DEGREE SIGN}")

                x2, y2 = (xy2 + np.array([50, -10])) + 50 * w1.unit
                plt.text(x2, y2, f"{theta2:.1f}\N{DEGREE SIGN}")
            else:
                xy1 = w1.top_coord  # - w1.cross_section.H / 2 * self.chord.unit
                xy2 = w2.top_coord  # + w2.cross_section.H / 2 * self.chord.unit

                if theta0 < 90:
                    arc1 = Arc(xy1, 100, 100, theta1=theta0 + 180,
                               theta2=theta0 + 180 + theta1)
                    arc2 = Arc(xy2, 100, 100, theta1=360 - theta0 - theta2,
                               theta2=theta0)
                else:
                    arc1 = Arc(xy1, 100, 100, theta1=theta0,
                               theta2=theta0 + theta1)
                    arc2 = Arc(xy2, 100, 100, theta1=180 + theta0 - theta2,
                               theta2=180 + theta0)

                x1, y1 = (xy1 + np.array([-50, -10])) - 50 * w1.unit
                plt.text(x1, y1, f"{theta1:.1f}\N{DEGREE SIGN}")

                x2, y2 = (xy2 + np.array([50, -10])) - 50 * w1.unit
                plt.text(x2, y2, f"{theta2:.1f}\N{DEGREE SIGN}")

            ax = plt.subplot(111)
            ax.add_patch(arc1)
            ax.add_patch(arc2)

        # NODES
        for key in self.nodes.keys():
            node = self.nodes[key]
            x, y = node.coord
            plt.scatter(x, y, s=100, c='pink')
            plt.text(x, y, str(key), horizontalalignment='center',
                     verticalalignment='center', fontsize=15)
            if x < X:
                x -= length
            elif x > X:
                x += length / 1.5
            if y < Y:
                y -= length
            elif y > Y:
                y += length

            # plt.text(x, y+0.08, f'Fx {node.Fx:.0f} kN', fontsize=11)
            # plt.text(x, y+0.05, f'Fy {node.Fy:.0f} kN', fontsize=11)
            # plt.text(x, y+0.02, f'Mz {node.Mz:.0f} kNm', fontsize=11)

        x, y = self.nodal_coordinates['c0'] - self.e * self.chord.perpendicular
        plt.scatter(x, y, s=100, c='pink')
        plt.text(x, y, "ec", horizontalalignment='center',
                 verticalalignment='center', fontsize=15)

        # ECCENTRICITY ELEMENTS
        for elem in self.elements.values():
            n1, n2 = elem.nodes
            x1, y1 = n1.coord
            x2, y2 = n2.coord
            plt.plot([x1, x2], [y1, y2], c='b')

        plt.axis('equal')
        plt.title("Joint " + str(self.jid))
        if show:
            plt.show()

    def update_forces(self):
        """
        Updates RHS_joint's forces
        """
        self.rhs_joint.M0 = self.M0
        self.rhs_joint.N0 = self.N0
        self.rhs_joint.V0 = self.V0

        if self.joint_type == 'Y':
            self.angle = np.degrees(self.theta1)

        elif self.joint_type == 'K':
            self.rhs_joint.angles = [np.degrees(self.theta1),
                                     np.degrees(self.theta2)]


        elif self.joint_type == 'KT':
            # TODO
            pass

def line_intersection(coordinates1, coordinates2):
    """
    Calculates coordinate where two members intersect
    :param coordinates: array of two arrays of two float values, [[x1,y1],[x2,y2]]
    :return [px, py]: array of two float values, coordinates for intersection
    source: https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection
    """
    start_node, end_node = coordinates1
    x1, y1 = start_node
    x2, y2 = end_node
    x3, y3 = coordinates2[0]
    x4, y4 = coordinates2[1]

    if ((x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)) == 0:
        return None
    else:
        px = ((x1 * y2 - y1 * x2) * (x3 - x4) - (x1 - x2) * (
                x3 * y4 - y3 * x4)) / (
                     (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4))
        py = ((x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (
                x3 * y4 - y3 * x4)) / (
                     (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4))

        return [round(px, 5), round(py, 5)]


if __name__ == '__main__':

    """
    from metku.frame2d.frame2d import *

    n_bays = 1
    n_storeys = 1
    L_bay = 10000
    H_storey = 5000

    simple_frame = [n_storeys, n_bays, H_storey, L_bay]

    frame = Frame2D(simple=simple_frame, create_beams=False, supports='fixed')

    simple_truss = dict(
        H0=H_storey,
        H1=1500,
        H2=2000,
        L1=L_bay / 2,
        H3=1500,
        dx=1000,
        n=16
    )
    """
    truss = PlaneTruss(span=10500, left_height=1000, mid_height=1150, right_height=1000,
                       num_elements=4)

    
    for tc in truss.top_chords:
        truss.add(LineLoad(tc, [-30, -30], 'y',load_id=LoadIDs.ULS))
    
    
    truss.add(XYHingedSupport(truss.top_coords[0]))
    truss.add(YHingedSupport(truss.top_coords[-1]))
    
    truss.plot()
    #truss.generate()
    #truss.calculate()
    #truss.bmd(scale=100,loads=False)

    # truss.H1 = 1000
    # truss.H2 = 1500
    # truss.H3 = 1300
    # truss.dx = 100
    #
    # truss.f.draw()

    """
    joint = truss.joints[4]
    w1, w2 = joint.webs.values()

    print("Before: ")
    print('e0', joint.e)
    print('g0', joint.g1)
    print('e1', w1.j1.e)
    print('g1', w1.j1.g1)
    print('e2', w2.j1.e)
    print('g2', w2.j1.g1)

    print("theta1", np.degrees(joint.theta1))
    print("theta2", np.degrees(joint.theta2))
    print("-" * 20)
    val = 50
    print("Changed g0 to:", val)
    print("-" * 20)
    joint.g1 = val
    print("After: ")
    print('e0', joint.e)
    print('g0', joint.g1)
    print('e1', w1.j1.e)
    print('g1', w1.j1.g1)
    print('e2', w2.j1.e)
    print('g2', w2.j1.g1)
    print("theta1", np.degrees(joint.theta1))
    print("theta2", np.degrees(joint.theta2))
    """


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Calculation methods for sandwich panels with steel faces

Created on Wed Aug  7 09:15:25 2019

@author: kmela
"""

import math
from metku.materials.steel_data import Steel
from metku.eurocodes.en1993.constants import gammaM2

fasteners = {'SXC14-S19-5.5':{'d':5.5,'dS':5.1,'d1':4.59}}

def ruukki_panel(name='SPA E', thickness=100, length=6000):
    
    if name == 'SPA E':
        TopFace = Face(tF=0.6,material="S280GD")
        BottomFace = Face(tF=0.5,material="S280GD")
        core = Core(tC=thickness-0.6-0.5,shear_modulus=3.7,density=110e-9)
        width = 1200
    
    
    panel = SandwichPanel([TopFace,BottomFace],core,width,length)
    
    return panel

class SandwichScrew:
    """ Class for sandwich panel fasteners """
    
    def __init__(self,d=5.5,dS=5.1,d1=4.59):
        """
        Constructor

        Parameters
        ----------
        d : float, optional
            Diameter [mm]. The default is 5.5.
        dS : TYPE, optional
            Diameter [mm]. The default is 5.1.
        d1 : TYPE, optional
            Diameter [mm]. The default is 4.59.

        Returns
        -------
        None.

        """
        
        self.d = d
        self.dS = dS
        self.d1 = d1
    

class Face:
    """ Face of sandwich panel """
    
    def __init__(self,tF,material="S280GD",tcoat=0.04):
        """ Constructor

            Parameters:
            -----------
                :param tF: thickness of face [mm]
                :param material: face material [mm]
                :param ctcoat: thickness of coating [mm]

                Variables:
                ----------
                :ivar t: thickness of face
                :ivar mat: material
                :ivar tcoat: coating thickness
                :ivar fy: yield strength [MPa]
                :ivar fu: ultimate strength [MPa]
        """
        
        self.t = tF
        self.mat = Steel(material)
        self.tcoat = tcoat        
        
    @property
    def fy(self):
        return self.mat.fy
    
    @property
    def fu(self):
        return self.mat.fu
    
    @property
    def E(self):
        return self.mat.E
    
    @property
    def tcor(self):
        """ Thickness of face core """
        return self.t-self.tcoat
    
    def A(self,b):
        """ Cross-sectional area of the face 
            :param b: width of panel        
        """
        
        return self.t*b
    
    def I(self,b):
        """ Second moment of area of the face with respect to its own
            neutral axis
            :param b: width of panel
        """
        return 1/12*b*self.t**3
       
    def EI(self,b):
        """ Bending stiffness of the face with respect ot its own neutral axis
        """
        return self.E*self.I(b)
    
    def EA(self,b):
        """ Axial stiffness of the face """
        return self.E*self.A(b)
            
class Core:
    """ Core material """

    def __init__(self,tC,shear_modulus,density):
        
        self.t = tC
        self.G = shear_modulus
        self.density = density
        
        

class SandwichPanel:
    """ Class for sandwich panels """
    
    def __init__(self,faces,core,width=1200,length=6000,c=[1000],
                 fastener=SandwichScrew(d=5.5,dS=5.1,d1=4.59)):
        """ Constructor

            Parameters:
            -----------
                :param faces: list of class Face objects
                :param core: class core object
                :param width, length: self-evident [mm]
                :param c: list of screw pair distances [mm], starting from
                        outermost screws

                Variables:
                ----------
                :ivar t: thickness of face
                :ivar mat: material
                :ivar tcoat: coating thickness
                :ivar fy: yield strength [MPa]
                :ivar fu: ultimate strength [MPa]
        """
        self.faces = faces
        self.core = core
        self.width = width
        self.length = length
        self.c = c
        self.fastener = fastener
        
    @property
    def thickness(self):
        """ Total thickness of panel """
        return self.faces[0].t + self.faces[1].t + self.core.t
        
    @property
    def d(self):
        """ distance between center lines of faces """
        return 0.5 * self.faces[0].t + 0.5 * self.faces[1].t + self.core.t

    @property
    def Atop(self):        
        return self.faces[0].A(self.width)

    @property
    def Abottom(self):        
        return self.faces[1].A(self.width)

    @property
    def S(self):
        """ Shear stiffness [N] """
        return self.core.G*self.width*self.d**2/self.core.t
    
    @property
    def EIo(self):
        """ Bending stiffness of faces """
        b = self.width
        return self.faces[0].EI(b) + self.faces[1].EI(b)
    
    @property
    def EIs(self):
        """ 'sandwich part' of the bending stiffness """
        EA = [face.EA(self.width) for face in self.faces]
        
        return EA[0]*EA[1]/sum(EA)*self.d**2
    
    @property
    def tF2_cor(self):
        """ Thickness of steel core of face 2 """
        
        return self.faces[1].tcor
    
    @property
    def fu_F2(self):
        """ Ultimate strength of face 2 """
        return self.faces[1].fu

    def fastener_transverse_stiffness(self,tsup,verb=False):
        """ Transverse stiffness of the fastener according to
            ECCS recommendations
            
            :param tsup: thickness of supporting structure [mm]
        """
        d1 = self.fastener.d1
        dS = self.fastener.dS
        
        # bending stiffness of fastener
        EI = 200000*math.pi*dS**4/64
        
        # stiffness of clamping in the supporting structure
        Csup = 2400*math.sqrt(tsup*d1**5)
        
        # Stiffness of internal face (hole elongation)
        if self.faces[1].tcor < 0.7:
            kF2 = 6.93*self.faces[1].fu*math.sqrt(self.faces[1].tcor**3*d1)/(0.26 + 0.8 * self.faces[1].t)
        else:
            kF2 = 4.2*self.faces[1].fu*math.sqrt(self.faces[1].tcor**3*d1)/0.373

        
        D = self.thickness
        xF = 1 - (1/kF2-0.5*D*tsup/Csup - D*tsup**2/8/EI)/(1/kF2+D**2/Csup+D**2*(2*D+3*tsup)/6/EI)
        
        kv = 1/(xF/kF2 + (tsup**2 + 2*(1-xF)*D*tsup)/4/Csup + (3*(1-xF)*D*tsup**2 + tsup**3)/24/EI)
        
        if verb:
            print("EI = {0:4.2f}".format(EI))
            print("Csup = {0:4.2f}".format(Csup))
            print("kF2 = {0:4.2f}".format(kF2))
            print("xF = {0:4.4f}".format(xF))
            
            print("1/k1 = {0:4.4g}".format(xF/kF2))
            print("1/k2 = {0:4.4g}".format((tsup**2 + 2*(1-xF*D*tsup))/4/Csup))
            print("1/k3 = {0:4.4g}".format((3*(1-xF)*D*tsup**2 + tsup**3)/24/EI))
            
        return kv
    
    def shear_stiffness_for_support(self,tsup,L,n):
        """
        Shear stiffness of the panel that it provides for stabilisation of an individual member

        Returns
        -------
        None.

        """
        kv = self.fastener_transverse_stiffness(tsup)
        
        Si = 0.5*kv*n/L*sum([ci**2 for ci in self.c])
        
        return Si
        
    def fastener_shear_resistance(self):
        """
        Shear resistance of fastener

        Returns
        -------
        None.

        """
        VRk = 4.2*math.sqrt(self.tF2_cor**3*self.fastener.d1)*self.fu_F2
        
        return VRk/gammaM2
        
    
def example():
    
    TopFace = Face(tF=0.6,material="S280GD")
    BottomFace = Face(tF=0.5,material="S280GD")
    core = Core(tC=200-0.6-0.5,shear_modulus=3.7,density=110e-9)
    
    panel = SandwichPanel([TopFace,BottomFace],core,
                           width=1200,length=6000)

    # Tuulikuorma
    qw = 0.6 # kN/m2
    
    qU = 1.5*qw*panel.width*1e-3
    MEd = qU*panel.length**2/8
    VEd = qU*panel.length/2
    d = panel.d
    
    sF1 = MEd/d/panel.Atop
    sF2 = MEd/d/panel.Abottom
    tauC = VEd/d/panel.width
    Ls = 100
    sC = VEd/Ls/panel.width
    
    print("Kuorma murtorajatilassa qU = {0:4.2f} kN/m".format(qU))
    print("Taivutusmomentti MEd = {0:4.2f} kNm".format(MEd*1e-6))
    print("Taivutusmomentti VEd = {0:4.2f} kN".format(VEd*1e-3))
    print("d = {0:4.2f} mm".format(d))
    print("Atop = {0:4.2f} mm2".format(panel.Atop))
    print("Abottom = {0:4.2f} mm2".format(panel.Abottom))
    print("sF1 = {0:4.2f} MPa".format(sF1))
    print("sF2 = {0:4.2f} MPa".format(sF2))
    print("tauC = {0:4.3f} MPa".format(tauC))
    print("SC = {0:4.3f} MPa".format(sC))

    return panel

if __name__ == '__main__':
    
    #f = Face(0.5,material="S250GD")
            
    p = example()
    
    #TopFace = Face(tF=0.5,material="S280GD")
    #BottomFace = Face(tF=0.5,material="S280GD")
    #Core = Core(tC=97,shear_modulus=3.7,density=110e-9)
    
    #SPA100 = SandwichPanel([TopFace,BottomFace],Core,
    #                       width=1200,length=6000)
    
    #kv = SPA100.fastener_transverse_stiffness(tsup=8)    
    #print("kv = {0:4.2f}".format(kv))